Require Export InfrastructureOperations.
Require Export InfrastructureLemmasTerms.
Require Export DeclarationTyping.
Set Implicit Arguments.

(******************************************************************************)
(* Context extension                                                          *)
(******************************************************************************)

Lemma extend_append (Γ : Env) (Δ1 Δ2 : Ext) :
  extend Γ (append Δ1 Δ2) = extend (extend Γ Δ1) Δ2.
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.

Lemma lengthExt_append (Δ1 Δ2 : Ext) :
  lengthExt (append Δ1 Δ2) = lengthExt Δ2 + lengthExt Δ1.
Proof.
  induction Δ2; simpl; auto.
Qed.

Lemma tshiftExt_append (c : nat) (Δ1 Δ2 : Ext) :
  tshiftExt c (append Δ1 Δ2) = append (tshiftExt c Δ1) (tshiftExt c Δ2).
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.
Hint Rewrite tshiftExt_append : interaction.

Lemma tsubstExt_append (X : nat) (S : Ty) (Δ1 Δ2 : Ext) :
  tsubstExt X S (append Δ1 Δ2) = append (tsubstExt X S Δ1) (tsubstExt X S Δ2).
Proof.
  induction Δ2; simpl; f_equal; auto.
Qed.
Hint Rewrite tsubstExt_append : interaction.

(******************************************************************************)
(* Well-formed context lemmas                                                 *)
(******************************************************************************)

Inductive wf_env : Env → Prop :=
  | wf_empty                                      : wf_env empty
  | wf_evar {Γ T}  (wΓ: wf_env Γ) (wT: wfTy Γ T) : wf_env (evar Γ T)
  | wf_etvar {Γ T} (wΓ: wf_env Γ) (wT: wfTy Γ T) : wf_env (etvar Γ T).
Hint Constructors wf_env.

(******************************************************************************)
(* Context lookups.                                                           *)
(******************************************************************************)

Lemma lookup_etvar_functional {Γ : Env} {X} {S T} :
  lookup_etvar Γ X S → lookup_etvar Γ X T → S = T.
Proof.
  intros gbS gbT; revert S gbS.
  induction gbT; inversion 1; subst; simpl; f_equal; auto.
Qed.

Lemma lookup_evar_functional {Γ : Env} {X} {S T} :
  lookup_evar Γ X S → lookup_evar Γ X T → S = T.
Proof.
  intros gvS gvT; revert S gvS.
  induction gvT; inversion 1; subst; simpl; f_equal; auto.
Qed.

(******************************************************************************)
(* Context insertions.                                                        *)
(******************************************************************************)

Inductive shift_etvar : nat → Env → Env → Prop :=
  | shift_etvar_here {Γ T} :
      shift_etvar 0 Γ (etvar Γ T)
  | shift_etvar_there_evar {Γ1 Γ2 c T} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar c (evar Γ1 T) (evar Γ2 (tshiftTy c T))
  | shift_etvar_there_etvar {Γ1 Γ2 c T} :
      shift_etvar c Γ1 Γ2 →
      shift_etvar (S c) (etvar Γ1 T) (etvar Γ2 (tshiftTy c T)).
Hint Constructors shift_etvar.

Lemma weaken_shift_etvar {Γ1 : Env} {Γ2 : Env} {Δ : Ext} {c} :
  shift_etvar c Γ1 Γ2 →
  shift_etvar c (extend Γ1 Δ) (extend Γ2 (tshiftExt c Δ)).
Proof.
  revert c Γ2; induction Δ; simpl; intros; try constructor; auto.
Qed.
Hint Resolve weaken_shift_etvar.

Inductive shift_evar : nat → Env → Env → Prop :=
  | shift_evar_here {Γ T} :
      shift_evar 0 Γ (evar Γ T)
  | shift_evar_there_evar {Γ1 Γ2 c T} :
      shift_evar c Γ1 Γ2 →
      shift_evar (S c) (evar Γ1 T) (evar Γ2 T)
  | shift_evar_there_etvar {Γ1 Γ2 c T} :
      shift_evar c Γ1 Γ2 →
      shift_evar c (etvar Γ1 T) (etvar Γ2 T).
Hint Constructors shift_evar.

Lemma weaken_shift_evar {Γ1 : Env} {Γ2 : Env} (Δ : Ext) {c} :
  shift_evar c Γ1 Γ2 →
  shift_evar (lengthExt Δ + c) (extend Γ1 Δ) (extend Γ2 Δ).
Proof.
  revert c Γ2; induction Δ; simpl; intros; try constructor; auto.
Qed.
Hint Resolve weaken_shift_evar.

(******************************************************************************)
(* Lemmas about shifting and context lookups.                                 *)
(******************************************************************************)

Lemma shift_etvar_lookup_etvar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ X T, lookup_etvar Γ1 X T → lookup_etvar Γ2 (shiftIndex c X) (tshiftTy c T).
Proof.
  induction ib; inversion 1; subst; simpl; try rewrite tshiftTy_comm; eauto.
Qed.
Hint Resolve shift_etvar_lookup_etvar.

Lemma shift_etvar_lookup_evar {Γ1 Γ2 c} (ib : shift_etvar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T → lookup_evar Γ2 x (tshiftTy c T).
Proof.
  induction ib; inversion 1; subst; simpl; try rewrite tshiftTy_comm; eauto.
Qed.
Hint Resolve shift_etvar_lookup_evar.

Lemma shift_evar_lookup_etvar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ X T, lookup_etvar Γ1 X T → lookup_etvar Γ2 X T.
Proof.
  induction iv; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve shift_evar_lookup_etvar.

Lemma shift_evar_lookup_evar {Γ1 Γ2 c} (iv : shift_evar c Γ1 Γ2) :
  ∀ x T, lookup_evar Γ1 x T → lookup_evar Γ2 (shiftIndex c x) T.
Proof.
  induction iv; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve shift_evar_lookup_evar.

(******************************************************************************)
(* Weakening lemmas for well-formedness                                       *)
(******************************************************************************)

Lemma shift_etvar_wftyp {Γ1 T} (wfT: wfTy Γ1 T) :
  ∀ {c Γ2}, shift_etvar c Γ1 Γ2 → wfTy Γ2 (tshiftTy c T).
Proof.
  induction wfT; simpl; econstructor; eauto.
Qed.
Hint Resolve shift_etvar_wftyp.

Lemma shift_evar_wftyp {Γ1 T} (wfT: wfTy Γ1 T) :
  ∀ {c Γ2}, shift_evar c Γ1 Γ2 → wfTy Γ2 T.
Proof.
  induction wfT; simpl; econstructor; eauto.
Qed.
Hint Resolve shift_evar_wftyp.

(******************************************************************************)
(* Strengthening lemmas for well-formedness                                   *)
(******************************************************************************)

Lemma compatible_context_wfty {Γ1 Γ2}
  (ξ: ∀ {X T1}, lookup_etvar Γ1 X T1 → ∃ T2, lookup_etvar Γ2 X T2) :
  ∀ T, wfTy Γ1 T → wfTy Γ2 T.
Proof.
  intros T wT; revert Γ2 ξ.
  induction wT; simpl; intros; eauto.
  - destruct (ξ _ _ H); eauto.
  - econstructor; eauto.
    apply IHwT2.
    inversion 1; subst; simpl.
    + eauto.
    + destruct (ξ _ _ H4); eauto.
Qed.

Lemma replace_etvar_wfty {Γ B1 B2 T} :
  wfTy (etvar Γ B1) T → wfTy (etvar Γ B2) T.
Proof.
  apply compatible_context_wfty; inversion 1; eauto.
Qed.

Lemma strengthen_evar_wfty {Γ S T} :
  wfTy (evar Γ S) T → wfTy Γ T.
Proof.
  apply compatible_context_wfty; inversion 1; eauto.
Qed.

Lemma strengthen_extend_wfty {Γ Δ T} :
  wfTy (extend Γ Δ) T → wfTy Γ T.
Proof.
  apply compatible_context_wfty; clear.
  induction Δ; simpl; inversion 1; eauto.
Qed.

(******************************************************************************)
(* Context etvar substitution                                                 *)
(******************************************************************************)

Inductive subst_etvar (S: Ty) : nat → Env → Env → Prop :=
  | tse_here {Γ B} :
      wfTy Γ S → Sub Γ S B →
      subst_etvar S 0 (etvar Γ B) Γ
  | tse_evar {X Γ1 Γ2 T} :
      subst_etvar S X Γ1 Γ2 →
      subst_etvar S X (evar Γ1 T) (evar Γ2 (tsubstTy X S T))
  | tse_etvar {Γ1 Γ2 X T} :
      subst_etvar S X Γ1 Γ2 →
      subst_etvar S (1 + X) (etvar Γ1 T) (etvar Γ2 (tsubstTy X S T)).
Hint Constructors subst_etvar.

Lemma weaken_subst_etvar {S} (Δ : Ext) :
  ∀ {x Γ1 Γ2},
    subst_etvar S x Γ1 Γ2 →
    subst_etvar S x (extend Γ1 Δ) (extend Γ2 (tsubstExt x S Δ)).
Proof.
  induction Δ; simpl; intros; try constructor; auto.
Qed.
Hint Resolve weaken_subst_etvar.

Lemma subst_etvar_lookup_etvar_wfty {S X Γ1 Γ2} (esub: subst_etvar S X Γ1 Γ2) :
  ∀ {Y T}, lookup_etvar Γ1 Y T → wfTy Γ2 (tsubstIndex X S Y).
Proof.
  induction esub; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve subst_etvar_lookup_etvar_wfty.

Lemma subst_etvar_wfty {S Γ1 T} (wfT: wfTy Γ1 T) :
  ∀ {X Γ2}, subst_etvar S X Γ1 Γ2 → wfTy Γ2 (tsubstTy X S T).
Proof.
  induction wfT; simpl; intros; try econstructor; eauto.
Qed.
Hint Resolve subst_etvar_wfty.

Lemma subst_etvar_wf_env {S X Γ1 Γ2} (esub: subst_etvar S X Γ1 Γ2) :
  wf_env Γ1 → wf_env Γ2.
Proof.
  induction esub; inversion 1; subst; simpl; eauto.
Qed.

Lemma subst_etvar_lookup_evar {S X Γ1 Γ2} (esub: subst_etvar S X Γ1 Γ2) :
  ∀ {Y T}, lookup_evar Γ1 Y T → lookup_evar Γ2 Y (tsubstTy X S T).
Proof.
  induction esub; inversion 1; subst; simpl;
    try rewrite tsubstTy_tshiftTy_reflection;
    try rewrite tsubstTy_tshiftTy_comm; try econstructor; eauto.
Qed.
Hint Resolve subst_etvar_lookup_evar.

(******************************************************************************)
(* Context evar substitution                                                  *)
(******************************************************************************)

Inductive subst_evar (s: Tm) : Subst → Env → Env → Prop :=
  | se_here {Γ S} :
      Typing Γ s S → subst_evar s sub_here (evar Γ S) Γ
  | se_evar {x Γ1 Γ2 T} :
      subst_evar s x Γ1 Γ2 →
      subst_evar s (sub_var x) (evar Γ1 T) (evar Γ2 T)
  | se_etvar {x Γ1 Γ2 T} :
      subst_evar s x Γ1 Γ2 →
      subst_evar s (sub_bound x) (etvar Γ1 T) (etvar Γ2 T).
Hint Constructors subst_evar.

Lemma weaken_subst_evar {s} (Δ : Ext) :
  ∀ {x Γ1 Γ2},
    subst_evar s x Γ1 Γ2 →
    subst_evar s (weaken_subst x (lengthExt Δ)) (extend Γ1 Δ) (extend Γ2 Δ).
Proof.
  induction Δ; simpl; intros; try constructor; auto.
Qed.
Hint Resolve weaken_subst_evar.

Lemma subst_evar_lookup_etvar {s X Γ1 Γ2} (esub: subst_evar s X Γ1 Γ2) :
  ∀ {Y T}, lookup_etvar Γ1 Y T → lookup_etvar Γ2 Y T.
Proof.
  induction esub; inversion 1; subst; simpl; eauto.
Qed.
Hint Resolve subst_evar_lookup_etvar.

Lemma subst_evar_wfty {s Γ1 T} (wfT: wfTy Γ1 T) :
  ∀ {X Γ2}, subst_evar s X Γ1 Γ2 → wfTy Γ2 T.
Proof.
  induction wfT; simpl; intros; econstructor; eauto.
Qed.
Hint Resolve subst_evar_wfty.

(******************************************************************************)
(* Forward reasoning about well-formedness                                    *)
(******************************************************************************)

Lemma lookup_etvar_wf {Γ} (wΓ: wf_env Γ) :
  ∀ {X T}, lookup_etvar Γ X T → wfTy Γ T.
Proof.
  induction wΓ; inversion 1; subst; simpl; eauto.
Qed.

Lemma lookup_evar_wf {Γ} (wΓ: wf_env Γ) :
  ∀ {X T}, lookup_evar Γ X T → wfTy Γ T.
Proof.
  induction wΓ; inversion 1; subst; simpl; eauto.
Qed.

Hint Extern 2 (wf_env _ _) =>
  match goal with
    | wG: wf_env (etvar _ _)                |- _ => inversion_clear wG; subst
    | wG: wf_env (evar _ _)                 |- _ => inversion_clear wG; subst
  end.

Hint Extern 2 (wfTy _ _) =>
  match goal with
    | wG: wf_env (etvar _ _)                |- _ => inversion_clear wG; subst
    | wG: wf_env (evar _ _)                 |- _ => inversion_clear wG; subst
    | wG: wf_env ?G, L: lookup_etvar ?G _ _ |- _ => apply (lookup_etvar_wf wG) in L
    | wG: wf_env ?G, L: lookup_evar ?G _ _  |- _ => apply (lookup_evar_wf wG) in L
    | w:  wfTy _ (tvar _)                  |- _ => inversion_clear w; subst
    | w:  wfTy _ top                       |- _ => inversion_clear w; subst
    | w:  wfTy _ (tarr _ _)                |- _ => inversion_clear w; subst
    | w:  wfTy _ (tall _ _)                |- _ => inversion_clear w; subst
    | H: wf_env ?G, IH : wf_env ?G -> _     |- _ => specialize (IH H); destruct_conjs
  end.
