Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.

Module EqSimplifierSymbolic (Spec : Specification).

  Module DEBRUIJN := DEBRUIJN Spec.
  Export DEBRUIJN.

  (* We want to proof the correctness of simplifying symbolic terms for sort
     terms, indices, cutoffs, hvl... For our purposes it's sufficient to model a
     terms without (meta-)variable binding and assume a fixed
     (meta-)context. *)

  (* Symbolic types and their interpretation. *)
  Inductive Symbolic :=
    | hvl
    | cutoff (α : Namespace)
    | index  (α : Namespace)
    | trace  (α : Namespace)
    | sort   (s : Sort).

  Definition Value (s : Symbolic) : Set :=
    match s with
      | hvl      => HVL
      | cutoff α => Cutoff α
      | index α  => Index α
      | trace α  => Trace α
      | sort s   => Term s
    end.

  (* Gamma is our fixed context. *)
  Context {Γ: list Symbolic}.

  (* For the purpose of evaluation we assume the assignment of concrete values
     to the context. Of course the lemmas don't depend on the choice and are
     universally quantified over this value environment. *)
  Variable (vs: ∀ s, s ∈ Γ → Value s).

  Section Expressions.

    (* Big mutually recursive list of expression for symbolic types. Each of
       these has a (meta-)variable case, cases for the constructors and core
       functions. *)
    Inductive ExpHVL : Set :=
      | ExHVar : hvl ∈ Γ → ExpHVL
      | ExH0
      | ExHS (α : Namespace) (exh: ExpHVL)
      | ExHAppend (h1 h2: ExpHVL)
      | ExHEvalFun {s} (fn: AuxFn s) (ext: ExpTerm s)
    with ExpCutoff : Namespace → Set :=
      | ExCVar {α} : cutoff α ∈ Γ → ExpCutoff α
      | ExC0 {α} : ExpCutoff α
      | ExCS {α} : ExpCutoff α → ExpCutoff α
      | ExCWeaken {α} : ExpCutoff α → ExpHVL → ExpCutoff α
    with ExpTrace : Namespace → Set :=
      | ExTVar {α} : trace α ∈ Γ → ExpTrace α
      | ExT0 {α} : ExpTrace α
      | ExTS {α β} : β ∈ depsOf (sortOf α) → ExpTrace α → ExpTrace α
      | ExTWeaken {α} : ExpTrace α → ExpHVL → ExpTrace α
    with ExpIndex : Namespace → Set :=
      | ExIVar {α} : index α ∈ Γ → ExpIndex α
      | ExI0 {α} : ExpIndex α
      | ExIS {α} : ExpIndex α → ExpIndex α
      | ExIWeaken {α} : ExpIndex α → ExpHVL → ExpIndex α
    with ExpTerm : Sort → Set :=
      | ExVar {s} : sort s ∈ Γ → ExpTerm s
      | ExCtorVar {α} : ExpIndex α → ExpTerm (sortOf α)
      | ExCtorFix {ct} : (∀ {s}, s ∈ ctorFields ct → ExpTerm s) →
          ExpTerm (ctorTarget ct)
      | ExShift {α s} (αIn: α ∈ depsOf s) : ExpCutoff α → ExpTerm s → ExpTerm s
      | ExSubst {α s} (αIn: α ∈ depsOf s) :
          ExpTrace α → ExpTerm (sortOf α) → ExpTerm s → ExpTerm s
      | ExSubstIndex {α} :
          ExpTrace α → ExpTerm (sortOf α) → ExpIndex α → ExpTerm (sortOf α).

    (* The following set of mutually recursive functions is an evaluator for the
       expressions that simply constructs values using the real constructor and
       calls the real functions. *)
    Fixpoint evalExpHVL (exh : ExpHVL) : HVL :=
      match exh with
        | ExHVar x          => vs _ x
        | ExH0              => HV0
        | ExHS α h          => HVS α (evalExpHVL h)
        | ExHAppend h1 h2   => append (evalExpHVL h1) (evalExpHVL h2)
        | ExHEvalFun s fn t => evalFunCall (evalExpTerm t) fn
      end
    with evalExpCutoff {α : Namespace} (exc : ExpCutoff α) : Cutoff α :=
      match exc with
        | ExCVar α x      => vs _ x
        | ExC0 α          => C0
        | ExCS α c        => CS (evalExpCutoff c)
        | ExCWeaken α c h => weaken (evalExpCutoff c) (evalExpHVL h)
      end
    with evalExpTrace {α : Namespace} (exx : ExpTrace α) : Trace α :=
      match exx with
        | ExTVar α x      => vs _ x
        | ExT0 α          => T0
        | ExTS α β βIn x  => TS βIn (evalExpTrace x)
        | ExTWeaken α x h => weaken (evalExpTrace x) (evalExpHVL h)
      end
    with evalExpIndex {α : Namespace} (ext : ExpIndex α) : Index α :=
      match ext with
        | ExIVar α x      => vs _ x
        | ExI0 α          => I0
        | ExIS α i        => IS (evalExpIndex i)
        | ExIWeaken α i h => weaken (evalExpIndex i) (evalExpHVL h)
      end
    with evalExpTerm {s : Sort} (ext : ExpTerm s) : Term s :=
      match ext with
        | ExVar s x               => vs _ x
        | ExCtorVar α i           => Var (evalExpIndex i)
        | ExCtorFix ct ts         => Fix (λ s sIn, evalExpTerm (ts s sIn))
        | ExShift α s αIn c t     => shift αIn (evalExpCutoff c) (evalExpTerm t)
        | ExSubst α s αIn xα tα t => subst αIn (evalExpTrace xα)
                                       (evalExpTerm tα) (evalExpTerm t)
        | ExSubstIndex α x t i    => substIndex (evalExpTrace x)
                                       (evalExpTerm t) (evalExpIndex i)
      end.

  End Expressions.

  Section Equality.

    Inductive eqexphvl : Set :=
      | eqexh_refl (exh : ExpHVL)
      | eqexh_sym (eq : eqexphvl)
      | eqexh_trans (eq1 : eqexphvl) (eq2 : eqexphvl)
      | eqexh_cong_succ (α: Namespace) (eq : eqexphvl)
      | eqexh_cong_append (eq1 : eqexphvl) (eq2 : eqexphvl)
      | eqexh_shift_fun {α s} (fn : AuxFn s) (exc: ExpCutoff α) (ext: ExpTerm s).

    Inductive EQEXH : eqexphvl → ExpHVL → ExpHVL → Prop :=
      | EQEXH_refl {h : ExpHVL} :
          EQEXH (eqexh_refl h) h h
      | EQEXH_cong_succ {α e h h'} :
          EQEXH e h h' →
          EQEXH (eqexh_cong_succ α e) (ExHS α h) (ExHS α h')
      | EQEXH_cong_append {e1 e2 h1 h1' h2 h2'} :
          EQEXH e1 h1 h1' →
          EQEXH e2 h2 h2' →
          EQEXH (eqexh_cong_append e1 e2) (ExHAppend h1 h2) (ExHAppend h1' h2')
      | EQEXH_shift_fun {α s} (αIn : α ∈ depsOf s) (fn : AuxFn s) exc ext :
          EQEXH
                (eqexh_shift_fun fn exc ext)
                (ExHEvalFun fn (ExShift αIn exc ext))
                (ExHEvalFun fn ext)
      | EQEXH_sym {e h1 h2} :
          EQEXH e h1 h2 →
          EQEXH (eqexh_sym e) h2 h1
      | EQEXH_trans {e1 e2 h1 h2 h3} :
          EQEXH e1 h1 h2 →
          EQEXH e2 h2 h3 →
          EQEXH (eqexh_trans e1 e2) h1 h3.

    Lemma EQEXH_sound :
      ∀ {eq h1 h2}, EQEXH eq h1 h2 → evalExpHVL h1 = evalExpHVL h2.
    Proof.
      induction 1; simpl; try congruence.
      apply evalFunCall_shift.
    Qed.

    Ltac crush_existT_eq :=
      repeat match goal with
        | H: existT ?P ?p _ = existT ?P ?p _ |- _ =>
          apply inj_pairT2 in H; subst
      end.

    Ltac crush_eqh :=
      simpl; eauto;
      repeat
        (match goal with
           | IH: EQEXH ?e _ _
             |- EQEXH (match ?e with _ => _ end) _ _ =>
             remember e; inversion IH; repeat subst
         end;
         crush_existT_eq;
         try (repeat econstructor; eauto; fail)
        );
        first
          [ econstructor;
            [ econstructor; eauto
            | repeat match goal with
                | H: _ = ?r |- EQHVL _ ?r _ =>
                  rewrite <- H
              end; econstructor; eauto
            ]
          | idtac
          ].

    Fixpoint simplifyHvlAppend (h1 h2 : ExpHVL) : ExpHVL :=
      match h2 with
        | ExH0          => h1
        | ExHS α h2     => ExHS α (simplifyHvlAppend h1 h2)
        | h2            => ExHAppend h1 h2
      end.

    Lemma simplifyHvlAppend_sound {h1 h2} :
      evalExpHVL (simplifyHvlAppend h1 h2) = append (evalExpHVL h1) (evalExpHVL h2).
    Proof.
      induction h2; simpl; congruence.
    Qed.

    Fixpoint simplifyHvl (h : ExpHVL) : ExpHVL :=
      match h with
        | ExHVar hIn        => ExHVar hIn
        | ExH0              => ExH0
        | ExHS α h          => ExHS α (simplifyHvl h)
        | ExHAppend h1 h2   => simplifyHvlAppend (simplifyHvl h1) (simplifyHvl h2)
        | ExHEvalFun s fn t => ExHEvalFun fn t
      end.

    Lemma simplifyHvl_sound {h} :
      evalExpHVL (simplifyHvl h) = evalExpHVL h.
    Proof.
      induction h; simpl; try congruence.
      rewrite simplifyHvlAppend_sound; congruence.
    Qed.

    Fixpoint simplify_eqexh_trans (e1 e2: eqexphvl) : eqexphvl :=
      match e1, e2 with
        | eqexh_refl _        , e2           => e2
        | e1                  , eqexh_refl _ => e1
        | eqexh_trans e11 e12 , e2           => simplify_eqexh_trans e11 (simplify_eqexh_trans e12 e2)
        | e1                  , e2           => eqexh_trans e1 e2
      end.

    Lemma simplify_eqexh_trans_sound {e1 h1 h2} (eq1: EQEXH e1 h1 h2) :
      ∀ {e2 h3}, EQEXH e2 h2 h3 → EQEXH (simplify_eqexh_trans e1 e2) h1 h3.
    Proof.
      induction eq1; intros ? ? eq2; crush_eqh.
    Qed.

    Definition simplify_eqexh_cong_succ α (e: eqexphvl) : eqexphvl :=
      match  e with
        | eqexh_refl h => eqexh_refl (ExHS α h)
        | e            => eqexh_cong_succ α e
      end.

    Lemma simplify_eqexh_cong_succ_sound {α e h1 h2} :
      EQEXH e h1 h2 →
      EQEXH (simplify_eqexh_cong_succ α e) (ExHS α h1) (ExHS α h2).
    Proof.
      intros eq; unfold simplify_eqexh_cong_succ; crush_eqh.
    Qed.

    Definition simplify_eqexh_cong_append (e1 e2: eqexphvl) : eqexphvl :=
      match e1, e2 with
        | eqexh_refl h1 , eqexh_refl h2 => eqexh_refl (ExHAppend h1 h2)
        | e1            , e2            => eqexh_cong_append e1 e2
      end.

    Lemma simplify_eqexh_cong_append_sound {e1 e2: eqexphvl} {h1 h1' h2 h2'} :
      EQEXH e1 h1 h1' →
      EQEXH e2 h2 h2' →
      EQEXH (simplify_eqexh_cong_append e1 e2) (ExHAppend h1 h2) (ExHAppend h1' h2').
    Proof.
      intros eq1 eq2; unfold simplify_eqexh_cong_append; crush_eqh.
    Qed.

    Fixpoint simplify_eqexh (eqh: eqexphvl) : eqexphvl :=
      match eqh with
        | eqexh_refl h1              => eqexh_refl h1
        | eqexh_sym e                => simplify_eqexh_sym e
        | eqexh_trans e1 e2          => simplify_eqexh_trans
                                          (simplify_eqexh e1) (simplify_eqexh e2)
        | eqexh_cong_succ α e        => simplify_eqexh_cong_succ α
                                          (simplify_eqexh e)
        | eqexh_cong_append e1 e2    => simplify_eqexh_cong_append
                                          (simplify_eqexh e1) (simplify_eqexh e2)
        | eqexh_shift_fun α s fn c t => eqexh_shift_fun fn c t
      end
    with simplify_eqexh_sym (eqh: eqexphvl) : eqexphvl :=
      match eqh with
        | eqexh_refl h1              => eqexh_refl h1
        | eqexh_sym e                => simplify_eqexh e
        | eqexh_trans e1 e2          => simplify_eqexh_trans
                                          (simplify_eqexh_sym e2) (simplify_eqexh_sym e1)
        | eqexh_cong_succ α e        => simplify_eqexh_cong_succ α
                                          (simplify_eqexh_sym e)
        | eqexh_cong_append e1 e2    => simplify_eqexh_cong_append
                                          (simplify_eqexh_sym e1) (simplify_eqexh_sym e2)
        | eqexh_shift_fun α s fn c t => eqexh_sym (eqexh_shift_fun fn c t)
      end.

    Lemma simplify_eqexh_sound_ind {e h1 h2} :
      EQEXH e h1 h2 →
      EQEXH (simplify_eqexh e) h1 h2 /\
      EQEXH (simplify_eqexh_sym e) h2 h1.
    Proof.
      intro eq; induction eq; simpl; split; destruct_conjs;
      eauto; try (repeat constructor; fail); crush_eqh;
      eauto using
        simplify_eqexh_cong_succ_sound,
        simplify_eqexh_cong_append_sound,
        simplify_eqexh_trans_sound.
    Qed.

    Lemma simplify_eqexh_sound {e h1 h2} :
      EQEXH e h1 h2 →
      EQEXH (simplify_eqexh e) h1 h2.
    Proof.
      eapply simplify_eqexh_sound_ind.
    Qed.

    Lemma simplify_eqexh_sym_sound {e h1 h2} :
      EQEXH e h1 h2 →
      EQEXH (simplify_eqexh_sym e) h2 h1.
    Proof.
      eapply simplify_eqexh_sound_ind.
    Qed.

  End Equality.

End EqSimplifierSymbolic.