
Require Export Coq.Unicode.Utf8.
Require Import Coq.Program.Tactics.
Require Import Prelude.
Require Import Specification.
Require Import DeBruijn.
Require Import DeBruijnInteraction.

Module WellFormedRecursive (Spec : Specification).

  Module Interaction := Interaction Spec.
  Export Interaction.

  Fixpoint wfindex (h : HVL) {α} (i : Index α) : Prop :=
    match h with
      | HV0 => False
      | HVS b h =>
        if α ≟ b
        then match i with
               | I0 => True
               | IS i => wfindex h i
             end
        else wfindex h i
    end.

  Fixpoint wfterm {s} (h : HVL) (t : Term s) : Prop :=
    match t with
      | Var _ i   => wfindex h i
      | Fix cn ts => ∀ s (sΓ: s ∈ ctorFields cn),
                      wfterm (append h (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))) (ts s sΓ)
    end.

  Section DependentInductionBased.

    Lemma wfindex_shift_hom {α} (h1 h2 : HVL) :
      forall (i : Index α),
          wfindex (append h1 h2) i ->
          wfindex (append (HVS α h1) h2) (shiftIndex (weakenCutoff C0 h2) i).
    Proof.
      induction h2 as [|β h2]; nsimpl.
      - auto.
      - destruct (α ≟ β) as [|αNotβ]; nsimpl.
        + destruct i; [reflexivity|exact (IHh2 i H)].
        + auto.
    Qed.

    Lemma wfindex_shift_het {α β} (αβ: α ≢ β) (h1 h2 : HVL) :
      forall (y : Index β),
        wfindex (append h1 h2) y →
        wfindex (append (HVS α h1) h2) y.
    Proof.
      induction h2 as [|γ h2]; nsimpl.
      - auto.
      - ncrush.
        + destruct y; auto.
        + auto.
    Qed.

    Lemma wfterm_ap {h1 h1' h2 h2': HVL} {s} {t1 t1' t2 t2' : Term s}:
      wfterm h1 t1 -> (wfterm h1' t1' -> wfterm h2' t2') ->
      h1 = h1' -> h2 = h2' -> t1 = t1' -> t2 = t2' ->
      wfterm h2 t2.
    Proof.
      intros; subst; auto.
    Qed.

    Lemma wfterm_shift_ind {α} :
      (forall {s} (t : Term s),
         (∀ (h1 h2 : HVL),
            wfterm (append h1 h2) t ->
            wfterm (append (HVS α h1) h2) (shift' (weakenCutoff (C0 (α := α)) h2) t))).
    Proof.
      simpl; induction t; nsimpl.
      - unfold shift'; ncrush.
        + apply wfindex_shift_hom; auto.
        + apply wfindex_shift_het; auto.
        + apply wfindex_shift_het; auto.
      - unfold shift'; ncrush; intros z zΓ; specialize (H z zΓ); revert H;
          unfold shift'; ncrush; intro IH.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ zΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ zΓ))
            by (symmetry; apply evalbindspec_shift).
          rewrite append_assoc.
          rewrite weakenCutoff_append.
          apply IH.
          rewrite <- append_assoc.
          apply H0.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ zΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ zΓ))
            by (symmetry; apply evalbindspec_shift).
          rewrite append_assoc.
          apply IH.
          rewrite <- append_assoc.
          apply H0.
        + elimtype Empty_set; eapply namespaceDepTransFix'; eauto.
        + rewrite append_assoc.
          apply IH.
          rewrite <- append_assoc.
          apply H0.
    Qed.

    Lemma wfterm_shift0_in {s} (t : Term s) {α} (αIn: α ∈ depsOf s) (h1 : HVL) :
      wfterm h1 t -> wfterm (HVS α h1) (shift αIn C0 t).
    Proof.
      generalize (@wfterm_shift_ind α s t h1 HV0); unfold shift'; nsimpl; auto.
    Qed.

    Lemma wfterm_shift0_notin {s} (t : Term s) {α} (αNotIn: α ∉ depsOf s) (h1 : HVL) :
      wfterm h1 t -> wfterm (HVS α h1) t.
    Proof.
      generalize (@wfterm_shift_ind α s t h1 HV0); nsimpl; auto.
    Qed.

    Lemma wfindex_strengthen {α β} (αβ: α ≢ β) (h1 h2 : HVL) :
      forall (y : Index β),
        wfindex (append (HVS α h1) h2) y →
        wfindex (append h1 h2) y.
    Proof.
      induction h2; nsimpl.
      - auto.
      - ncrush.
        + destruct y; auto.
        + auto.
    Qed.

    Lemma wfterm_strengthen {α} :
      (forall {s} (t : Term s) (αNotIn: α ∉ depsOf s) (h1 h2 : HVL),
           wfterm (append (HVS α h1) h2) t ->
           wfterm (append h1 h2) t).
    Proof.
      induction t; nsimpl.
      - eapply wfindex_strengthen; eauto; nsimpl.
      - specialize (H s sΓ); rename H into IH.
        rewrite append_assoc.
        apply IH; ncrush.
        rewrite <- append_assoc.
        apply H0.
    Qed.

    Lemma wfindex_substIndex_hom {α} (h1 h2 : HVL) (s : Term (sortOf α)) :
      forall (i : Index α),
          wfterm h1 s →
          wfindex (append (HVS α h1) h2) i ->
          wfterm (append h1 h2) (substIndex (weakenTrace T0 h2) s i).
    Proof.
      induction h2; nsimpl.
      - destruct i; auto.
      - revert H0; ncrush; nsimpl.
        + destruct i; nsimpl.
          * auto.
          * apply wfterm_shift0_in; auto.
        + rewrite <- substIndex_shiftIndex0_het_comm1.
          destruct i; nsimpl.
          * apply wfterm_shift0_in; auto.
          * apply wfterm_shift0_in; auto.
          * nsimpl.
        + * apply wfterm_shift0_notin; auto.
    Qed.

    Lemma wfterm_subst {α} (h1: HVL) (tα: Term (sortOf α)) (wf_tα: wfterm h1 tα) :
      (forall {s} (t : Term s) (αIn: α ∈ depsOf s) (h2 : HVL),
           wfterm (append (HVS α h1) h2) t ->
           wfterm (append h1 h2) (subst αIn (weakenTrace T0 h2) tα t)).
    Proof.
      simpl; induction t; nsimpl.
      - intros; ncrush.
        + apply wfindex_substIndex_hom; auto.
        + eapply wfindex_strengthen; eauto.
      - ncrush; specialize (H s sΓ).
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
           by (symmetry; apply evalbindspec_subst).
          rewrite append_assoc.
          rewrite weakenTrace_append.
          apply H.
          rewrite <- append_assoc.
          apply H0.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_subst).
          rewrite append_assoc.
          apply (@wfterm_strengthen α); auto.
          rewrite <- append_assoc.
          apply H0.
    Qed.

    Lemma wfterm_subst0 {s} (t : Term s) {α} (αIn: α ∈ depsOf s)
          (h1 : HVL) (tα: Term (sortOf α)) :
      wfterm h1 tα → wfterm (HVS α h1) t -> wfterm h1 (subst αIn T0 tα t).
    Proof.
      intros wf_tα wf_t; generalize (wfterm_subst h1 tα wf_tα t αIn HV0); auto.
    Qed.

  End DependentInductionBased.

  Section RelationBased.

    Inductive Insertion {α} : ∀ (c: Cutoff α) (h1 h2: HVL), Prop :=
      | ins_here {h} :
          Insertion C0 h (HVS α h)
      | ins_there_eq {c h1 h2} :
          Insertion c h1 h2 →
          Insertion (CS c) (HVS α h1) (HVS α h2)
      | ins_there_neq {c h1 h2 β} (αNotβ: α ≢ β) :
          Insertion c h1 h2 →
          Insertion c (HVS β h1) (HVS β h2).

    Lemma weaken_insertion {α} :
      ∀ (h h1 h2 : HVL) (c : Cutoff α),
        Insertion c h1 h2 →
        Insertion (weakenCutoff c h) (append h1 h) (append h2 h).
    Proof.
      induction h; nsimpl.
      - auto.
      - ncrush.
        + constructor; eauto.
        + constructor; eauto.
    Qed.

    Lemma wfindex_shift'_hom {α} {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ i, wfindex h1 i ->
           wfindex h2 (shiftIndex c i).
    Proof.
      induction ins; nsimpl.
      - auto.
      - destruct i; simpl.
        + constructor.
        + apply IHins; auto.
      - apply IHins; auto.
    Qed.

    Lemma wfindex_shift'_het {α β} (αNotβ: α ≢ β) {c: Cutoff α} {h1 h2} (ins: Insertion c h1 h2) :
      ∀ (i : Index β),
        wfindex h1 i -> wfindex h2 i.
    Proof.
      induction ins; nsimpl.
      - auto.
      - auto.
      - revert H; ncrush; nsimpl.
        + destruct i; simpl; auto.
        + auto.
    Qed.

    Lemma wfterm_shift'_ind {α} :
      (forall {s} (t : Term s) (h1 : HVL) (wf_t: wfterm h1 t),
       ∀ {c : Cutoff α} {h2}, Insertion c h1 h2 → wfterm h2 (shift' c t)).
    Proof.
      simpl; induction t; nsimpl.
      - unfold shift'; ncrush.
        + eapply wfindex_shift'_hom; eauto.
        + eapply wfindex_shift'_het; eauto.
        + eapply wfindex_shift'_het; eauto.
      - unfold shift'; ncrush;
          intros s sΓ; specialize (H s sΓ _ (wf_t s sΓ));
          revert H; unfold shift'; ncrush; intros H.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          apply H.
          apply weaken_insertion; auto.
        + replace (evalBindSpec _ HV0 (ctorBindspec _ _ sΓ))
          with (evalBindSpec ts HV0 (ctorBindspec _ _ sΓ))
            by (symmetry; apply evalbindspec_shift).
          eapply H; auto.
          apply weaken_insertion; eauto.
        + elimtype Empty_set; eapply namespaceDepTransFix'; eauto.
        + eapply H; eauto.
          apply weaken_insertion; eauto.
    Qed.

  End RelationBased.

End WellFormedRecursive.
