
Require Import Prelude.

(*  ____                  _  __ _           _   _                  *)
(* / ___| _ __   ___  ___(_)/ _(_) ___ __ _| |_(_) ___  _ __  ___  *)
(* \___ \| '_ \ / _ \/ __| | |_| |/ __/ _` | __| |/ _ \| '_ \/ __| *)
(*  ___) | |_) |  __/ (__| |  _| | (_| (_| | |_| | (_) | | | \__ \ *)
(* |____/| .__/ \___|\___|_|_| |_|\___\__,_|\__|_|\___/|_| |_|___/ *)
(*       |_|                                                       *)


Module Type Specification.

  (* We use types to represent the namespaces, constructors and *)
  (* auxiliary functions. One can construct these types given the list *)
  (* of defined namespaces from the specification using Σ-types: *)

  (*  Variable RawNamespaces : list string. *)
  (*  Definition Namespace := { α : string & α ∈ RawNamespaces }. *)

  (* One also needs uniqueness in RawNamespaces, we won't bother doing it. *)


  Parameter Namespace : Set.
  Parameter Sort : Set.

  (*         _   _                                                            *)
  (*        | \ | | __ _ _ __ ___   ___  ___ _ __   __ _  ___ ___  ___        *)
  (*        |  \| |/ _` | '_ ` _ \ / _ \/ __| '_ \ / _` |/ __/ _ \/ __|       *)
  (*        | |\  | (_| | | | | | |  __/\__ \ |_) | (_| | (_|  __/\__ \       *)
  (*        |_| \_|\__,_|_| |_| |_|\___||___/ .__/ \__,_|\___\___||___/       *)
  (*                                        |_|                               *)

  (* The sort that is associated with a namespace, i.e. the sort that
     has a variable constructor for that namespace *)
  Parameter sortOf : Namespace → Sort.

  (* We require namespaces labels to support decidable equality. *)
  Axiom seq_namespace_dec :
    ∀ (α β : Namespace), (α ≡ β) + (α ≢ β).

  Infix "≟"  := seq_namespace_dec          (at level 80, no associativity).
  Infix "∈?" := (In_dec seq_namespace_dec) (at level 80, no associativity).

  Parameter depsOf         : Sort → list Namespace.

  Axiom namespaceDepVar      : ∀ (α : Namespace), α ∈ depsOf (sortOf α).
  Axiom namespaceDepTransVar : ∀ {α β s},
                                    α ∈ depsOf (sortOf β) →
                                    β ∈ depsOf s →
                                    α ∈ depsOf s.

  (* We currently don't model distinctivness of depsOf (yet), i.e. that
     the list of namespaces for a sort doesn't contain duplicates. The only
     property that we really need, is uniqueness of containment proofs.
  *)
  Axiom namespaceDepIn :
    ∀ (α : Namespace) (s : Sort)
      (αIn1 αIn2 : α ∈ depsOf s),
        αIn1 = αIn2.

  Lemma seq_namespace_dec_refl :
    ∀ α, (α ≟ α) = inl seq_refl.
  Proof.
    intros; destruct (α ≟ α).
    dependent destruction s.
    reflexivity.
    destruct (n seq_refl).
  Qed.

  Corollary seq_namespace_dec_seq {α β : Namespace} :
    ∀ (αβ: α ≡ β), (α ≟ β) = inl αβ.
  Proof.
    intro αβ.
    destruct (α ≟ β); f_equal.
    destruct s.
    dependent destruction αβ.
    reflexivity.
    destruct (n αβ).
  Qed.

  Corollary seq_namespace_dec_nseq {α β : Namespace} :
    ∀ (αβ : α ≢ β), (α ≟ β) = inr αβ.
  Proof.
    intro αβ.
    destruct (α ≟ β); f_equal.
    destruct (αβ s).
    extensionality e.
    destruct (n e).
  Qed.

  Corollary namespaceDep_In
    (α : Namespace) (s : Sort) (αIn : α ∈ depsOf s) :
      (α ∈? depsOf s) = inl αIn.
  Proof.
    destruct (α ∈? depsOf s); f_equal.
    apply namespaceDepIn.
    destruct (n αIn).
  Qed.

  Corollary namespaceDep_notIn
    (α : Namespace) (s : Sort) (αNotIn : α ∉ depsOf s) :
      (α ∈? depsOf s) = inr αNotIn.
  Proof.
    destruct (α ∈? depsOf s); f_equal.
    destruct (αNotIn i).
    extensionality e.
    destruct (n e).
  Qed.

  (*                         ____             _                               *)
  (*                        / ___|  ___  _ __| |_ ___                         *)
  (*                        \___ \ / _ \| '__| __/ __|                        *)
  (*                         ___) | (_) | |  | |_\__ \                        *)
  (*                        |____/ \___/|_|   \__|___/                        *)

  (* The Set Ctor is intended to represent all non-variable constructor
     of the specification. We don't bother representing variable
     constructors explicitly: we will have exactly one variable constructor
     per namespaces that will construct a term of the appropriate sort. *)
  Parameter Ctor              : Set.

  (* The sorts of the fields of a constructor. Note that we will
     be using strongly typed de Bruijn indices into this list rather
     than field names for this development. Field names are thus
     completely erased.
  *)
  Parameter ctorFields        : Ctor → list Sort.
  Parameter ctorTarget        : Ctor → Sort.

  Axiom namespaceDepTransFix :
    ∀ {ct : Ctor} {t : Sort} (tIn: t ∈ ctorFields ct),
      depsOf t ⊆ depsOf (ctorTarget ct).

  Lemma namespaceDepTransFix' {ct : Ctor} {α} (αNotIn : α ∉ depsOf (ctorTarget ct)) :
    ∀ {s}, s ∈ ctorFields ct → α ∉ depsOf s.
  Proof.
    intros s s_ct  αIn.
    apply αNotIn.
    apply (namespaceDepTransFix s_ct); auto.
  Qed.

  (*             ____  _           _   ____                                   *)
  (*            | __ )(_)_ __   __| | / ___| _ __   ___  ___ ___              *)
  (*            |  _ \| | '_ \ / _` | \___ \| '_ \ / _ \/ __/ __|             *)
  (*            | |_) | | | | | (_| |  ___) | |_) |  __/ (__\__ \             *)
  (*            |____/|_|_| |_|\__,_| |____/| .__/ \___|\___|___/             *)
  (*                                        |_|                               *)

  (* AuxFn maps a Sort to a type representing the functions defined
     on that sort. *)
  Parameter AuxFn             : Sort → Set.

  (* This function assign to each auxiliary function |fn : s -> α|
     the range namespace α.
  *)
  Parameter auxFnsTarget : ∀ {s : Sort}, AuxFn s → list Namespace.
  (* Sorts with variables don't have auxiliary functions. We rather use
    a global variable for this than to put it into the well-formedness
    of every variable constructor like in the paper. *)
  Axiom auxFnsNoVar : ∀ {α : Namespace}, AuxFn (sortOf α) →
                                             forall (A:Type), A.

  Inductive VarListItem (Γ : list Sort) (αs : list Namespace) : Set :=
    | vlOne    {α} : α ∈ αs → VarListItem Γ αs
    | vlAuxFn  {s} (fn : AuxFn s) (sΓ : s ∈ Γ) :
        auxFnsTarget fn ⊆ αs →
        VarListItem Γ αs.

  Definition VarListExp (Γ : list Sort) αs : Set :=
    list (VarListItem Γ αs).

  (* This gives for every field of a constructor the appropriate
     binding specification. *)

  Parameter ctorBindspecNamespaces : ∀ ct s, s ∈ ctorFields ct → list Namespace.
  Parameter ctorBindspec : ∀ ct s (sIn: s ∈ ctorFields ct),
                             VarListExp (ctorFields ct) (ctorBindspecNamespaces ct s sIn).
  (* Gives the definition of an auxiliary |f| function for
     a conrete namespace α *)
  Parameter ctorAuxFn    : ∀ {ct} (fn : AuxFn (ctorTarget ct)),
                             VarListExp (ctorFields ct) (auxFnsTarget fn).

End Specification.
