Require Import Coq.Program.Equality.
Require Import Coq.Program.Tactics.
Require Export Needle.
Section Namespace.
  Inductive Namespace : Type :=
    | TmVar 
    | TyVar .
  Lemma eq_namespace_dec (a : Namespace) (b : Namespace) :
    (sumbool (a =
    b) (not (a =
    b))).
  Proof.
    decide equality .
  Defined.
End Namespace.

Section HVarlist.
  Inductive Hvl : Type :=
    | H0 
    | HS (a : Namespace) (k : Hvl).
  
  Fixpoint appendHvl (k : Hvl) (k0 : Hvl) {struct k0} : Hvl :=
    match k0 with
      | (H0) => k
      | (HS a k0) => (HS a (appendHvl k k0))
    end.
  
  Lemma appendHvl_assoc  :
    (forall (k : Hvl) (k0 : Hvl) (k1 : Hvl) ,
      ((appendHvl (appendHvl k k0) k1) =
      (appendHvl k (appendHvl k0 k1)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End HVarlist.

Section Index.
  Inductive Index (a : Namespace) : Type :=
    | I0 
    | IS : (Index a) -> (Index a).
  
  Global Arguments I0 [a] .
  Global Arguments IS [a] _ .
  
  Lemma eq_index_dec {a : Namespace} (i : (Index a)) (j : (Index a)) :
    (sumbool (i =
    j) (not (i =
    j))).
  Proof.
    decide equality .
  Qed.
  
  Fixpoint weakenIndexTmVar (x : (Index TmVar)) (k : Hvl) {struct k} : (Index TmVar) :=
    match k with
      | (H0) => x
      | (HS a k) => match a with
        | (TmVar) => (IS (weakenIndexTmVar x k))
        | _ => (weakenIndexTmVar x k)
      end
    end.
  Fixpoint weakenIndexTyVar (X : (Index TyVar)) (k : Hvl) {struct k} : (Index TyVar) :=
    match k with
      | (H0) => X
      | (HS a k) => match a with
        | (TyVar) => (IS (weakenIndexTyVar X k))
        | _ => (weakenIndexTyVar X k)
      end
    end.
  Lemma weakenIndexTmVar_append  :
    (forall (x : (Index TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTmVar (weakenIndexTmVar x k) k0) =
      (weakenIndexTmVar x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenIndexTyVar_append  :
    (forall (X : (Index TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenIndexTyVar (weakenIndexTyVar X k) k0) =
      (weakenIndexTyVar X (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
End Index.

Section Terms.
  Inductive Ty : Set :=
    | tvar (X : (Index TyVar))
    | top 
    | tarrow (T1 : Ty) (T2 : Ty)
    | tall (T1 : Ty) (T2 : Ty)
    | tprod (T1 : Ty) (T2 : Ty).
  Scheme ind_Ty := Induction for Ty Sort Prop.
  
  Inductive Pat : Set :=
    | pvar 
    | pprod (p1 : Pat) (p2 : Pat).
  Scheme ind_Pat := Induction for Pat Sort Prop.
  
  Inductive Tm : Set :=
    | var (x : (Index TmVar))
    | abs (T : Ty) (t : Tm)
    | app (t1 : Tm) (t2 : Tm)
    | tabs (T : Ty) (t : Tm)
    | tapp (t : Tm) (T : Ty)
    | prod (t1 : Tm) (t2 : Tm)
    | lett (p : Pat) (t1 : Tm)
        (t2 : Tm).
  Scheme ind_Tm := Induction for Tm Sort Prop.
End Terms.

Section Functions.
  Fixpoint bindPat (p : Pat) {struct p} : Hvl :=
    match p with
      | (pvar) => (appendHvl (HS TmVar H0) H0)
      | (pprod p1 p2) => (appendHvl (bindPat p1) (appendHvl (bindPat p2) H0))
    end.
  Scheme ind_bindPat_Pat := Induction for Pat Sort Prop.
End Functions.

Section Shift.
  Inductive Cutoff (a : Namespace) : Type :=
    | C0 
    | CS : (Cutoff a) -> (Cutoff a).
  
  Global Arguments C0 {a} .
  Global Arguments CS {a} _ .
  Fixpoint weakenCutoffTmVar (c : (Cutoff TmVar)) (k : Hvl) {struct k} : (Cutoff TmVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TmVar) => (CS (weakenCutoffTmVar c k))
        | _ => (weakenCutoffTmVar c k)
      end
    end.
  Fixpoint weakenCutoffTyVar (c : (Cutoff TyVar)) (k : Hvl) {struct k} : (Cutoff TyVar) :=
    match k with
      | (H0) => c
      | (HS a k) => match a with
        | (TyVar) => (CS (weakenCutoffTyVar c k))
        | _ => (weakenCutoffTyVar c k)
      end
    end.
  
  Lemma weakenCutoffTmVar_append  :
    (forall (c : (Cutoff TmVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTmVar (weakenCutoffTmVar c k) k0) =
      (weakenCutoffTmVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Lemma weakenCutoffTyVar_append  :
    (forall (c : (Cutoff TyVar)) (k : Hvl) (k0 : Hvl) ,
      ((weakenCutoffTyVar (weakenCutoffTyVar c k) k0) =
      (weakenCutoffTyVar c (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint shiftIndex (c : (Cutoff TmVar)) (x : (Index TmVar)) {struct c} : (Index TmVar) :=
    match c with
      | (C0) => (IS x)
      | (CS c) => match x with
        | (I0) => I0
        | (IS x) => (IS (shiftIndex c x))
      end
    end.
  Fixpoint tshiftIndex (c : (Cutoff TyVar)) (X : (Index TyVar)) {struct c} : (Index TyVar) :=
    match c with
      | (C0) => (IS X)
      | (CS c) => match X with
        | (I0) => I0
        | (IS X) => (IS (tshiftIndex c X))
      end
    end.
  Fixpoint tshiftTy (c : (Cutoff TyVar)) (S0 : Ty) {struct S0} : Ty :=
    match S0 with
      | (tvar X) => (tvar (tshiftIndex c X))
      | (top) => (top)
      | (tarrow T1 T2) => (tarrow (tshiftTy (weakenCutoffTyVar c H0) T1) (tshiftTy (weakenCutoffTyVar c H0) T2))
      | (tall T0 T3) => (tall (tshiftTy (weakenCutoffTyVar c H0) T0) (tshiftTy (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) T3))
      | (tprod T4 T5) => (tprod (tshiftTy (weakenCutoffTyVar c H0) T4) (tshiftTy (weakenCutoffTyVar c H0) T5))
    end.
  Fixpoint shiftTm (c : (Cutoff TmVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (var x) => (var (shiftIndex c x))
      | (abs T t) => (abs T (shiftTm (weakenCutoffTmVar c (appendHvl (HS TmVar H0) H0)) t))
      | (app t1 t2) => (app (shiftTm (weakenCutoffTmVar c H0) t1) (shiftTm (weakenCutoffTmVar c H0) t2))
      | (tabs T0 t0) => (tabs T0 (shiftTm (weakenCutoffTmVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (tapp t3 T1) => (tapp (shiftTm (weakenCutoffTmVar c H0) t3) T1)
      | (prod t4 t5) => (prod (shiftTm (weakenCutoffTmVar c H0) t4) (shiftTm (weakenCutoffTmVar c H0) t5))
      | (lett p t6 t7) => (lett p (shiftTm (weakenCutoffTmVar c H0) t6) (shiftTm (weakenCutoffTmVar c (appendHvl (bindPat p) H0)) t7))
    end.
  Fixpoint tshiftTm (c : (Cutoff TyVar)) (s : Tm) {struct s} : Tm :=
    match s with
      | (var x) => (var x)
      | (abs T t) => (abs (tshiftTy (weakenCutoffTyVar c H0) T) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TmVar H0) H0)) t))
      | (app t1 t2) => (app (tshiftTm (weakenCutoffTyVar c H0) t1) (tshiftTm (weakenCutoffTyVar c H0) t2))
      | (tabs T0 t0) => (tabs (tshiftTy (weakenCutoffTyVar c H0) T0) (tshiftTm (weakenCutoffTyVar c (appendHvl (HS TyVar H0) H0)) t0))
      | (tapp t3 T1) => (tapp (tshiftTm (weakenCutoffTyVar c H0) t3) (tshiftTy (weakenCutoffTyVar c H0) T1))
      | (prod t4 t5) => (prod (tshiftTm (weakenCutoffTyVar c H0) t4) (tshiftTm (weakenCutoffTyVar c H0) t5))
      | (lett p t6 t7) => (lett p (tshiftTm (weakenCutoffTyVar c H0) t6) (tshiftTm (weakenCutoffTyVar c (appendHvl (bindPat p) H0)) t7))
    end.
End Shift.

Section Weaken.
  Fixpoint weakenTy (S0 : Ty) (k : Hvl) {struct k} : Ty :=
    match k with
      | (H0) => S0
      | (HS TmVar k) => (weakenTy S0 k)
      | (HS TyVar k) => (tshiftTy C0 (weakenTy S0 k))
    end.
  Fixpoint weakenPat (p : Pat) (k : Hvl) {struct k} : Pat :=
    match k with
      | (H0) => p
      | (HS TmVar k) => (weakenPat p k)
      | (HS TyVar k) => (weakenPat p k)
    end.
  Fixpoint weakenTm (s : Tm) (k : Hvl) {struct k} : Tm :=
    match k with
      | (H0) => s
      | (HS TmVar k) => (shiftTm C0 (weakenTm s k))
      | (HS TyVar k) => (tshiftTm C0 (weakenTm s k))
    end.
End Weaken.

Section Subst.
  Inductive Trace (a : Namespace) : Type :=
    | X0 
    | XS (b : Namespace)
        (T : (Trace a)).
  
  Global Arguments X0 [a] .
  Global Arguments XS [a] b _ .
  Fixpoint weakenTrace {a : Namespace} (x : (Trace a)) (k : Hvl) {struct k} : (Trace a) :=
    match k with
      | (H0) => x
      | (HS b k) => (XS b (weakenTrace x k))
    end.
  Lemma weakenTrace_append (a : Namespace) :
    (forall (x : (Trace a)) (k : Hvl) (k0 : Hvl) ,
      ((weakenTrace (weakenTrace x k) k0) =
      (weakenTrace x (appendHvl k k0)))).
  Proof.
    needleGenericWeakenAppend .
  Qed.
  Fixpoint substIndex (d : (Trace TmVar)) (s : Tm) (x : (Index TmVar)) {struct d} : Tm :=
    match d with
      | (X0) => match x with
        | (I0) => s
        | (IS x) => (var x)
      end
      | (XS TmVar d) => match x with
        | (I0) => (var I0)
        | (IS x) => (shiftTm C0 (substIndex d s x))
      end
      | (XS TyVar d) => (tshiftTm C0 (substIndex d s x))
    end.
  Fixpoint tsubstIndex (d : (Trace TyVar)) (S0 : Ty) (X : (Index TyVar)) {struct d} : Ty :=
    match d with
      | (X0) => match X with
        | (I0) => S0
        | (IS X) => (tvar X)
      end
      | (XS TmVar d) => (tsubstIndex d S0 X)
      | (XS TyVar d) => match X with
        | (I0) => (tvar I0)
        | (IS X) => (tshiftTy C0 (tsubstIndex d S0 X))
      end
    end.
  Fixpoint tsubstTy (d : (Trace TyVar)) (S0 : Ty) (S1 : Ty) {struct S1} : Ty :=
    match S1 with
      | (tvar X) => (tsubstIndex d S0 X)
      | (top) => (top)
      | (tarrow T1 T2) => (tarrow (tsubstTy d S0 T1) (tsubstTy d S0 T2))
      | (tall T0 T3) => (tall (tsubstTy d S0 T0) (tsubstTy (XS TyVar d) S0 T3))
      | (tprod T4 T5) => (tprod (tsubstTy d S0 T4) (tsubstTy d S0 T5))
    end.
  Fixpoint substTm (d : (Trace TmVar)) (s : Tm) (s0 : Tm) {struct s0} : Tm :=
    match s0 with
      | (var x) => (substIndex d s x)
      | (abs T t) => (abs T (substTm (XS TmVar d) s t))
      | (app t1 t2) => (app (substTm d s t1) (substTm d s t2))
      | (tabs T0 t0) => (tabs T0 (substTm (XS TyVar d) s t0))
      | (tapp t3 T1) => (tapp (substTm d s t3) T1)
      | (prod t4 t5) => (prod (substTm d s t4) (substTm d s t5))
      | (lett p t6 t7) => (lett p (substTm d s t6) (substTm (weakenTrace d (bindPat p)) s t7))
    end.
  Fixpoint tsubstTm (d : (Trace TyVar)) (S0 : Ty) (s : Tm) {struct s} : Tm :=
    match s with
      | (var x) => (var x)
      | (abs T t) => (abs (tsubstTy d S0 T) (tsubstTm (XS TmVar d) S0 t))
      | (app t1 t2) => (app (tsubstTm d S0 t1) (tsubstTm d S0 t2))
      | (tabs T0 t0) => (tabs (tsubstTy d S0 T0) (tsubstTm (XS TyVar d) S0 t0))
      | (tapp t3 T1) => (tapp (tsubstTm d S0 t3) (tsubstTy d S0 T1))
      | (prod t4 t5) => (prod (tsubstTm d S0 t4) (tsubstTm d S0 t5))
      | (lett p t6 t7) => (lett p (tsubstTm d S0 t6) (tsubstTm (weakenTrace d (bindPat p)) S0 t7))
    end.
End Subst.

Global Hint Resolve (f_equal (shiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTm C0)) : cong_shift0.
Global Hint Resolve (f_equal (tshiftTy C0)) : cong_shift0.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : interaction.
 Hint Rewrite weakenCutoffTmVar_append weakenCutoffTyVar_append : weakencutoff_append.
 Hint Rewrite weakenTrace_append : interaction.
 Hint Rewrite weakenTrace_append : weakentrace_append.
Lemma stability_weaken_bindPat  :
  (forall (k : Hvl) (p : Pat) ,
    ((bindPat (weakenPat p k)) =
    (bindPat p))).
Proof.
  needleGenericStabilityWeaken .
Qed.
 Hint Rewrite stability_weaken_bindPat : interaction.
 Hint Rewrite stability_weaken_bindPat : stability_weaken.
Section SubstShiftReflection.
  Lemma substIndex0_shiftIndex0_reflection_ind (s : Tm) :
    (forall (k4 : Hvl) (x9 : (Index TmVar)) ,
      ((substIndex (weakenTrace X0 k4) s (shiftIndex (weakenCutoffTmVar C0 k4) x9)) =
      (var x9))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Lemma tsubstIndex0_tshiftIndex0_reflection_ind (S0 : Ty) :
    (forall (k4 : Hvl) (X8 : (Index TyVar)) ,
      ((tsubstIndex (weakenTrace X0 k4) S0 (tshiftIndex (weakenCutoffTyVar C0 k4) X8)) =
      (tvar X8))).
  Proof.
    needleGenericSubstIndexShiftIndexReflectionInd .
  Qed.
  Definition tsubst0_tshift0_Ty_reflection_ind := (ind_Ty (fun (S2 : Ty) =>
    (forall (k7 : Hvl) (S3 : Ty) ,
      ((tsubstTy (weakenTrace X0 k7) S3 (tshiftTy (weakenCutoffTyVar C0 k7) S2)) =
      S2))) (fun (X14 : (Index TyVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      (tsubstIndex0_tshiftIndex0_reflection_ind S2 k7 X14))) (fun (k7 : Hvl) (S2 : Ty) =>
    eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 tarrow (IHT6 k7 S2) (IHT7 k7 S2))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 tall (IHT6 k7 S2) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHT7 (HS TyVar k7) S2)))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 tprod (IHT6 k7 S2) (IHT7 k7 S2)))).
  Definition subst0_shift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (s1 : Tm) ,
      ((substTm (weakenTrace X0 k7) s1 (shiftTm (weakenCutoffTmVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (s0 : Tm) =>
      (substIndex0_shiftIndex0_reflection_ind s0 k7 x14))) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt53 (HS TmVar k7) s0)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 app (IHt53 k7 s0) (IHt54 k7 s0))) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 tabs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt53 (HS TyVar k7) s0)))) (fun (t : Tm) IHt53 (T : Ty) (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 tapp (IHt53 k7 s0) eq_refl)) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal2 prod (IHt53 k7 s0) (IHt54 k7 s0))) (fun (p : Pat) (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (s0 : Tm) =>
    (f_equal3 lett eq_refl (IHt53 k7 s0) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar X0 k7 (bindPat p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k7 (bindPat p)) eq_refl)) (IHt54 (appendHvl k7 (bindPat p)) s0))))).
  Definition tsubst0_tshift0_Tm_reflection_ind := (ind_Tm (fun (s0 : Tm) =>
    (forall (k7 : Hvl) (S2 : Ty) ,
      ((tsubstTm (weakenTrace X0 k7) S2 (tshiftTm (weakenCutoffTyVar C0 k7) s0)) =
      s0))) (fun (x14 : (Index TmVar)) =>
    (fun (k7 : Hvl) (S2 : Ty) =>
      eq_refl)) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 abs (let IHT := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT k7 S2)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k7 (HS TmVar H0)) eq_refl eq_refl) (IHt53 (HS TmVar k7) S2)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 app (IHt53 k7 S2) (IHt54 k7 S2))) (fun (T : Ty) (t : Tm) IHt53 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 tabs (let IHT := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT k7 S2)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k7 (HS TyVar H0)) eq_refl eq_refl) (IHt53 (HS TyVar k7) S2)))) (fun (t : Tm) IHt53 (T : Ty) (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 tapp (IHt53 k7 S2) (let IHT := (tsubst0_tshift0_Ty_reflection_ind T) in
    (IHT k7 S2)))) (fun (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (S2 : Ty) =>
    (f_equal2 prod (IHt53 k7 S2) (IHt54 k7 S2))) (fun (p : Pat) (t1 : Tm) IHt53 (t2 : Tm) IHt54 (k7 : Hvl) (S2 : Ty) =>
    (f_equal3 lett eq_refl (IHt53 k7 S2) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k7 (bindPat p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k7 (bindPat p)) eq_refl)) (IHt54 (appendHvl k7 (bindPat p)) S2))))).
  Definition tsubstTy0_tshiftTy0_reflection (S3 : Ty) : (forall (S2 : Ty) ,
    ((tsubstTy X0 S2 (tshiftTy C0 S3)) =
    S3)) := (tsubst0_tshift0_Ty_reflection_ind S3 H0).
  Definition substTm0_shiftTm0_reflection (s1 : Tm) : (forall (s0 : Tm) ,
    ((substTm X0 s0 (shiftTm C0 s1)) =
    s1)) := (subst0_shift0_Tm_reflection_ind s1 H0).
  Definition tsubstTm0_tshiftTm0_reflection (s0 : Tm) : (forall (S2 : Ty) ,
    ((tsubstTm X0 S2 (tshiftTm C0 s0)) =
    s0)) := (tsubst0_tshift0_Tm_reflection_ind s0 H0).
End SubstShiftReflection.
Section ShiftInteraction.
  Section ShiftIndexCommInd.
    Lemma shiftIndex_shiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TmVar)) (x : (Index TmVar)) ,
        ((shiftIndex (weakenCutoffTmVar (CS c) k) (shiftIndex (weakenCutoffTmVar C0 k) x)) =
        (shiftIndex (weakenCutoffTmVar C0 k) (shiftIndex (weakenCutoffTmVar c k) x)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
    Lemma tshiftIndex_tshiftIndex0_comm_ind  :
      (forall (k : Hvl) (c : (Cutoff TyVar)) (X : (Index TyVar)) ,
        ((tshiftIndex (weakenCutoffTyVar (CS c) k) (tshiftIndex (weakenCutoffTyVar C0 k) X)) =
        (tshiftIndex (weakenCutoffTyVar C0 k) (tshiftIndex (weakenCutoffTyVar c k) X)))).
    Proof.
      needleGenericShiftIndexCommInd .
    Qed.
  End ShiftIndexCommInd.
  Section ShiftCommInd.
    Definition tshift_tshift0_Ty_comm_ind := (ind_Ty (fun (S0 : Ty) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar (CS c4) k4) (tshiftTy (weakenCutoffTyVar C0 k4) S0)) =
        (tshiftTy (weakenCutoffTyVar C0 k4) (tshiftTy (weakenCutoffTyVar c4 k4) S0))))) (fun (X8 : (Index TyVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        (f_equal tvar (tshiftIndex_tshiftIndex0_comm_ind k4 c4 X8)))) (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tarrow (IHT6 k4 c4) (IHT7 k4 c4))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tall (IHT6 k4 c4) (IHT7 (HS TyVar k4) c4))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tprod (IHT6 k4 c4) (IHT7 k4 c4)))).
    Definition shift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar (CS c4) k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (shiftTm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        (f_equal var (shiftIndex_shiftIndex0_comm_ind k4 c4 x9)))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 abs eq_refl (IHt35 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 app (IHt35 k4 c4) (IHt36 k4 c4))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 tabs eq_refl (IHt35 (HS TyVar k4) c4))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 tapp (IHt35 k4 c4) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 prod (IHt35 k4 c4) (IHt36 k4 c4))) (fun (p : Pat) (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal3 lett eq_refl (IHt35 k4 c4) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append (CS c4) k4 (bindPat p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bindPat p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bindPat p)) c4) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k4 (bindPat p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c4 k4 (bindPat p))) eq_refl))))))).
    Definition shift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c4 k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (shiftTm (weakenCutoffTmVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 abs eq_refl (IHt35 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 app (IHt35 k4 c4) (IHt36 k4 c4))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 tabs eq_refl (IHt35 (HS TyVar k4) c4))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 tapp (IHt35 k4 c4) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal2 prod (IHt35 k4 c4) (IHt36 k4 c4))) (fun (p : Pat) (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TmVar)) =>
      (f_equal3 lett eq_refl (IHt35 k4 c4) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c4 k4 (bindPat p)) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bindPat p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bindPat p)) c4) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k4 (bindPat p))) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c4 k4 (bindPat p))) eq_refl))))))).
    Definition tshift_shift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar c4 k4) (shiftTm (weakenCutoffTmVar C0 k4) s)) =
        (shiftTm (weakenCutoffTmVar C0 k4) (tshiftTm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 abs eq_refl (IHt35 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 app (IHt35 k4 c4) (IHt36 k4 c4))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tabs eq_refl (IHt35 (HS TyVar k4) c4))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tapp (IHt35 k4 c4) eq_refl)) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 prod (IHt35 k4 c4) (IHt36 k4 c4))) (fun (p : Pat) (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal3 lett eq_refl (IHt35 k4 c4) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c4 k4 (bindPat p)) (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k4 (bindPat p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bindPat p)) c4) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k4 (bindPat p))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c4 k4 (bindPat p))) eq_refl))))))).
    Definition tshift_tshift0_Tm_comm_ind := (ind_Tm (fun (s : Tm) =>
      (forall (k4 : Hvl) (c4 : (Cutoff TyVar)) ,
        ((tshiftTm (weakenCutoffTyVar (CS c4) k4) (tshiftTm (weakenCutoffTyVar C0 k4) s)) =
        (tshiftTm (weakenCutoffTyVar C0 k4) (tshiftTm (weakenCutoffTyVar c4 k4) s))))) (fun (x9 : (Index TmVar)) =>
      (fun (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 abs (let IHT := (tshift_tshift0_Ty_comm_ind T) in
      (IHT k4 c4)) (IHt35 (HS TmVar k4) c4))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 app (IHt35 k4 c4) (IHt36 k4 c4))) (fun (T : Ty) (t : Tm) IHt35 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tabs (let IHT := (tshift_tshift0_Ty_comm_ind T) in
      (IHT k4 c4)) (IHt35 (HS TyVar k4) c4))) (fun (t : Tm) IHt35 (T : Ty) (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 tapp (IHt35 k4 c4) (let IHT := (tshift_tshift0_Ty_comm_ind T) in
      (IHT k4 c4)))) (fun (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal2 prod (IHt35 k4 c4) (IHt36 k4 c4))) (fun (p : Pat) (t1 : Tm) IHt35 (t2 : Tm) IHt36 (k4 : Hvl) (c4 : (Cutoff TyVar)) =>
      (f_equal3 lett eq_refl (IHt35 k4 c4) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append (CS c4) k4 (bindPat p)) (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k4 (bindPat p)) eq_refl)) (eq_trans (IHt36 (appendHvl k4 (bindPat p)) c4) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k4 (bindPat p))) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c4 k4 (bindPat p))) eq_refl))))))).
  End ShiftCommInd.
  Section ShiftComm.
    Definition tshift_tshift0_Ty_comm (S0 : Ty) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTy (CS c4) (tshiftTy C0 S0)) =
      (tshiftTy C0 (tshiftTy c4 S0)))) := (tshift_tshift0_Ty_comm_ind S0 H0).
    Definition shift_shift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shiftTm (CS c4) (shiftTm C0 s)) =
      (shiftTm C0 (shiftTm c4 s)))) := (shift_shift0_Tm_comm_ind s H0).
    Definition shift_tshift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TmVar)) ,
      ((shiftTm c4 (tshiftTm C0 s)) =
      (tshiftTm C0 (shiftTm c4 s)))) := (shift_tshift0_Tm_comm_ind s H0).
    Definition tshift_shift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTm c4 (shiftTm C0 s)) =
      (shiftTm C0 (tshiftTm c4 s)))) := (tshift_shift0_Tm_comm_ind s H0).
    Definition tshift_tshift0_Tm_comm (s : Tm) : (forall (c4 : (Cutoff TyVar)) ,
      ((tshiftTm (CS c4) (tshiftTm C0 s)) =
      (tshiftTm C0 (tshiftTm c4 s)))) := (tshift_tshift0_Tm_comm_ind s H0).
  End ShiftComm.
End ShiftInteraction.
 Hint Rewrite shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : interaction.
 Hint Rewrite shift_shift0_Tm_comm shift_tshift0_Tm_comm tshift_shift0_Tm_comm tshift_tshift0_Tm_comm tshift_tshift0_Ty_comm : shift_shift0.
Section WeakenShift.
  Lemma weakenTy_tshiftTy  :
    (forall (k60 : Hvl) (c9 : (Cutoff TyVar)) (S40 : Ty) ,
      ((weakenTy (tshiftTy c9 S40) k60) =
      (tshiftTy (weakenCutoffTyVar c9 k60) (weakenTy S40 k60)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_shiftTm  :
    (forall (k60 : Hvl) (c9 : (Cutoff TmVar)) (s18 : Tm) ,
      ((weakenTm (shiftTm c9 s18) k60) =
      (shiftTm (weakenCutoffTmVar c9 k60) (weakenTm s18 k60)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
  Lemma weakenTm_tshiftTm  :
    (forall (k60 : Hvl) (c9 : (Cutoff TyVar)) (s18 : Tm) ,
      ((weakenTm (tshiftTm c9 s18) k60) =
      (tshiftTm (weakenCutoffTyVar c9 k60) (weakenTm s18 k60)))).
  Proof.
    needleGenericWeakenShift .
  Qed.
End WeakenShift.
Section ShiftSubstInteraction.
  Section ShiftSubstIndexCommInd.
    Lemma shiftTm_substIndex0_comm_ind (c4 : (Cutoff TmVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((shiftTm (weakenCutoffTmVar c4 k7) (substIndex (weakenTrace X0 k7) s0 x14)) =
        (substIndex (weakenTrace X0 k7) (shiftTm c4 s0) (shiftIndex (weakenCutoffTmVar (CS c4) k7) x14)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTm_substIndex0_comm_ind (c4 : (Cutoff TyVar)) (s0 : Tm) :
      (forall (k7 : Hvl) (x14 : (Index TmVar)) ,
        ((tshiftTm (weakenCutoffTyVar c4 k7) (substIndex (weakenTrace X0 k7) s0 x14)) =
        (substIndex (weakenTrace X0 k7) (tshiftTm c4 s0) x14))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tshiftTy_tsubstIndex0_comm_ind (c4 : (Cutoff TyVar)) (S2 : Ty) :
      (forall (k7 : Hvl) (X14 : (Index TyVar)) ,
        ((tshiftTy (weakenCutoffTyVar c4 k7) (tsubstIndex (weakenTrace X0 k7) S2 X14)) =
        (tsubstIndex (weakenTrace X0 k7) (tshiftTy c4 S2) (tshiftIndex (weakenCutoffTyVar (CS c4) k7) X14)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End ShiftSubstIndexCommInd.
  Section ShiftSubstCommInd.
    Definition tshift_tsubst0_Ty_comm_ind := (ind_Ty (fun (S5 : Ty) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S6 : Ty) ,
        ((tshiftTy (weakenCutoffTyVar c9 k12) (tsubstTy (weakenTrace X0 k12) S6 S5)) =
        (tsubstTy (weakenTrace X0 k12) (tshiftTy c9 S6) (tshiftTy (weakenCutoffTyVar (CS c9) k12) S5))))) (fun (X21 : (Index TyVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        (tshiftTy_tsubstIndex0_comm_ind c9 S5 k12 X21))) (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 tarrow (IHT6 k12 c9 S5) (IHT7 k12 c9 S5))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 tall (IHT6 k12 c9 S5) (eq_trans (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT7 (HS TyVar k12) c9 S5) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 tprod (IHT6 k12 c9 S5) (IHT7 k12 c9 S5)))).
    Definition shift_subst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (s3 : Tm) ,
        ((shiftTm (weakenCutoffTmVar c9 k12) (substTm (weakenTrace X0 k12) s3 s2)) =
        (substTm (weakenTrace X0 k12) (shiftTm c9 s3) (shiftTm (weakenCutoffTmVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
        (shiftTm_substIndex0_comm_ind c9 s2 k12 x26))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 app (IHt89 k12 c9 s2) (IHt90 k12 c9 s2))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 tapp (IHt89 k12 c9 s2) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal2 prod (IHt89 k12 c9 s2) (IHt90 k12 c9 s2))) (fun (p : Pat) (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (s2 : Tm) =>
      (f_equal3 lett eq_refl (IHt89 k12 c9 s2) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c9 k12 (bindPat p)) (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bindPat p)) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (bindPat p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append (CS c9) k12 (bindPat p))) eq_refl))))))).
    Definition shift_tsubst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) ,
        ((shiftTm (weakenCutoffTmVar c9 k12) (tsubstTm (weakenTrace X0 k12) S5 s2)) =
        (tsubstTm (weakenTrace X0 k12) S5 (shiftTm (weakenCutoffTmVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 app (IHt89 k12 c9 S5) (IHt90 k12 c9 S5))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 tapp (IHt89 k12 c9 S5) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal2 prod (IHt89 k12 c9 S5) (IHt90 k12 c9 S5))) (fun (p : Pat) (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TmVar)) (S5 : Ty) =>
      (f_equal3 lett eq_refl (IHt89 k12 c9 S5) (eq_trans (f_equal2 shiftTm (weakenCutoffTmVar_append c9 k12 (bindPat p)) (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bindPat p)) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (bindPat p))) eq_refl (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append c9 k12 (bindPat p))) eq_refl))))))).
    Definition tshift_subst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (s3 : Tm) ,
        ((tshiftTm (weakenCutoffTyVar c9 k12) (substTm (weakenTrace X0 k12) s3 s2)) =
        (substTm (weakenTrace X0 k12) (tshiftTm c9 s3) (tshiftTm (weakenCutoffTyVar c9 k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
        (tshiftTm_substIndex0_comm_ind c9 s2 k12 x26))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 app (IHt89 k12 c9 s2) (IHt90 k12 c9 s2))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 tapp (IHt89 k12 c9 s2) eq_refl)) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal2 prod (IHt89 k12 c9 s2) (IHt90 k12 c9 s2))) (fun (p : Pat) (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (s2 : Tm) =>
      (f_equal3 lett eq_refl (IHt89 k12 c9 s2) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c9 k12 (bindPat p)) (f_equal3 substTm (weakenTrace_append TmVar X0 k12 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bindPat p)) c9 s2) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k12 (bindPat p))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append c9 k12 (bindPat p))) eq_refl))))))).
    Definition tshift_tsubst0_Tm_comm_ind := (ind_Tm (fun (s2 : Tm) =>
      (forall (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) ,
        ((tshiftTm (weakenCutoffTyVar c9 k12) (tsubstTm (weakenTrace X0 k12) S5 s2)) =
        (tsubstTm (weakenTrace X0 k12) (tshiftTy c9 S5) (tshiftTm (weakenCutoffTyVar (CS c9) k12) s2))))) (fun (x26 : (Index TmVar)) =>
      (fun (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 abs (let IHT := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT k12 c9 S5)) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TmVar k12) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 app (IHt89 k12 c9 S5) (IHt90 k12 c9 S5))) (fun (T : Ty) (t : Tm) IHt89 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 tabs (let IHT := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT k12 c9 S5)) (eq_trans (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt89 (HS TyVar k12) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt89 (T : Ty) (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 tapp (IHt89 k12 c9 S5) (let IHT := (tshift_tsubst0_Ty_comm_ind T) in
      (IHT k12 c9 S5)))) (fun (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal2 prod (IHt89 k12 c9 S5) (IHt90 k12 c9 S5))) (fun (p : Pat) (t1 : Tm) IHt89 (t2 : Tm) IHt90 (k12 : Hvl) (c9 : (Cutoff TyVar)) (S5 : Ty) =>
      (f_equal3 lett eq_refl (IHt89 k12 c9 S5) (eq_trans (f_equal2 tshiftTm (weakenCutoffTyVar_append c9 k12 (bindPat p)) (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k12 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt90 (appendHvl k12 (bindPat p)) c9 S5) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k12 (bindPat p))) eq_refl (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append (CS c9) k12 (bindPat p))) eq_refl))))))).
  End ShiftSubstCommInd.
  Section ShiftSubstComm.
    Definition tshiftTy_tsubstTy0_comm (S6 : Ty) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((tshiftTy c9 (tsubstTy X0 S5 S6)) =
      (tsubstTy X0 (tshiftTy c9 S5) (tshiftTy (CS c9) S6)))) := (tshift_tsubst0_Ty_comm_ind S6 H0).
    Definition shiftTm_substTm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TmVar)) (s2 : Tm) ,
      ((shiftTm c9 (substTm X0 s2 s3)) =
      (substTm X0 (shiftTm c9 s2) (shiftTm (CS c9) s3)))) := (shift_subst0_Tm_comm_ind s3 H0).
    Definition shiftTm_tsubstTm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TmVar)) (S5 : Ty) ,
      ((shiftTm c9 (tsubstTm X0 S5 s2)) =
      (tsubstTm X0 S5 (shiftTm c9 s2)))) := (shift_tsubst0_Tm_comm_ind s2 H0).
    Definition tshiftTm_substTm0_comm (s3 : Tm) : (forall (c9 : (Cutoff TyVar)) (s2 : Tm) ,
      ((tshiftTm c9 (substTm X0 s2 s3)) =
      (substTm X0 (tshiftTm c9 s2) (tshiftTm c9 s3)))) := (tshift_subst0_Tm_comm_ind s3 H0).
    Definition tshiftTm_tsubstTm0_comm (s2 : Tm) : (forall (c9 : (Cutoff TyVar)) (S5 : Ty) ,
      ((tshiftTm c9 (tsubstTm X0 S5 s2)) =
      (tsubstTm X0 (tshiftTy c9 S5) (tshiftTm (CS c9) s2)))) := (tshift_tsubst0_Tm_comm_ind s2 H0).
  End ShiftSubstComm.
  Section SubstShiftIndexCommInd.
    Lemma substIndex_shiftTm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TmVar d) k12) s2 (shiftIndex (weakenCutoffTmVar C0 k12) x26)) =
        (shiftTm (weakenCutoffTmVar C0 k12) (substIndex (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma substIndex_tshiftTm0_comm_ind (d : (Trace TmVar)) (s2 : Tm) :
      (forall (k12 : Hvl) (x26 : (Index TmVar)) ,
        ((substIndex (weakenTrace (XS TyVar d) k12) s2 x26) =
        (tshiftTm (weakenCutoffTyVar C0 k12) (substIndex (weakenTrace d k12) s2 x26)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_shiftTy0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (X21 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TmVar d) k12) S5 X21) =
        (tsubstIndex (weakenTrace d k12) S5 X21))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
    Lemma tsubstIndex_tshiftTy0_comm_ind (d : (Trace TyVar)) (S5 : Ty) :
      (forall (k12 : Hvl) (X21 : (Index TyVar)) ,
        ((tsubstIndex (weakenTrace (XS TyVar d) k12) S5 (tshiftIndex (weakenCutoffTyVar C0 k12) X21)) =
        (tshiftTy (weakenCutoffTyVar C0 k12) (tsubstIndex (weakenTrace d k12) S5 X21)))).
    Proof.
      needleGenericShiftSubstIndexCommInd .
    Qed.
  End SubstShiftIndexCommInd.
  Section SubstShiftCommInd.
    Definition tsubst_tshift0_Ty_comm_ind := (ind_Ty (fun (S24 : Ty) =>
      (forall (k45 : Hvl) (d32 : (Trace TyVar)) (S25 : Ty) ,
        ((tsubstTy (weakenTrace (XS TyVar d32) k45) S25 (tshiftTy (weakenCutoffTyVar C0 k45) S24)) =
        (tshiftTy (weakenCutoffTyVar C0 k45) (tsubstTy (weakenTrace d32 k45) S25 S24))))) (fun (X28 : (Index TyVar)) =>
      (fun (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
        (tsubstIndex_tshiftTy0_comm_ind d32 S24 k45 X28))) (fun (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tarrow (IHT6 k45 d32 S24) (IHT7 k45 d32 S24))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tall (IHT6 k45 d32 S24) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TyVar d32) k45 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT7 (HS TyVar k45) d32 S24) (f_equal2 tshiftTy eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d32 k45 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tprod (IHT6 k45 d32 S24) (IHT7 k45 d32 S24)))).
    Definition subst_shift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k45 : Hvl) (d32 : (Trace TmVar)) (s17 : Tm) ,
        ((substTm (weakenTrace (XS TmVar d32) k45) s17 (shiftTm (weakenCutoffTmVar C0 k45) s16)) =
        (shiftTm (weakenCutoffTmVar C0 k45) (substTm (weakenTrace d32 k45) s17 s16))))) (fun (x38 : (Index TmVar)) =>
      (fun (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
        (substIndex_shiftTm0_comm_ind d32 s16 k45 x38))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d32) k45 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k45) d32 s16) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 app (IHt125 k45 d32 s16) (IHt126 k45 d32 s16))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d32) k45 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k45) d32 s16) (f_equal2 shiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 tapp (IHt125 k45 d32 s16) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 prod (IHt125 k45 d32 s16) (IHt126 k45 d32 s16))) (fun (p : Pat) (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal3 lett eq_refl (IHt125 k45 d32 s16) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TmVar d32) k45 (bindPat p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k45 (bindPat p)) eq_refl)) (eq_trans (IHt126 (appendHvl k45 (bindPat p)) d32 s16) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k45 (bindPat p))) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (bindPat p))) eq_refl eq_refl))))))).
    Definition subst_tshift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k45 : Hvl) (d32 : (Trace TmVar)) (s17 : Tm) ,
        ((substTm (weakenTrace (XS TyVar d32) k45) s17 (tshiftTm (weakenCutoffTyVar C0 k45) s16)) =
        (tshiftTm (weakenCutoffTyVar C0 k45) (substTm (weakenTrace d32 k45) s17 s16))))) (fun (x38 : (Index TmVar)) =>
      (fun (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
        (substIndex_tshiftTm0_comm_ind d32 s16 k45 x38))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d32) k45 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k45) d32 s16) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 app (IHt125 k45 d32 s16) (IHt126 k45 d32 s16))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d32) k45 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k45) d32 s16) (f_equal2 tshiftTm eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 tapp (IHt125 k45 d32 s16) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal2 prod (IHt125 k45 d32 s16) (IHt126 k45 d32 s16))) (fun (p : Pat) (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TmVar)) (s16 : Tm) =>
      (f_equal3 lett eq_refl (IHt125 k45 d32 s16) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar (XS TyVar d32) k45 (bindPat p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k45 (bindPat p)) eq_refl)) (eq_trans (IHt126 (appendHvl k45 (bindPat p)) d32 s16) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k45 (bindPat p))) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar d32 k45 (bindPat p))) eq_refl eq_refl))))))).
    Definition tsubst_shift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) ,
        ((tsubstTm (weakenTrace d32 k45) S24 (shiftTm (weakenCutoffTmVar C0 k45) s16)) =
        (shiftTm (weakenCutoffTmVar C0 k45) (tsubstTm (weakenTrace d32 k45) S24 s16))))) (fun (x38 : (Index TmVar)) =>
      (fun (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d32 k45 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k45) d32 S24) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 app (IHt125 k45 d32 S24) (IHt126 k45 d32 S24))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d32 k45 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k45) d32 S24) (f_equal2 shiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tapp (IHt125 k45 d32 S24) eq_refl)) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 prod (IHt125 k45 d32 S24) (IHt126 k45 d32 S24))) (fun (p : Pat) (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal3 lett eq_refl (IHt125 k45 d32 S24) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d32 k45 (bindPat p)) eq_refl (f_equal2 shiftTm (weakenCutoffTmVar_append C0 k45 (bindPat p)) eq_refl)) (eq_trans (IHt126 (appendHvl k45 (bindPat p)) d32 S24) (f_equal2 shiftTm (eq_sym (weakenCutoffTmVar_append C0 k45 (bindPat p))) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (bindPat p))) eq_refl eq_refl))))))).
    Definition tsubst_tshift0_Tm_comm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) ,
        ((tsubstTm (weakenTrace (XS TyVar d32) k45) S24 (tshiftTm (weakenCutoffTyVar C0 k45) s16)) =
        (tshiftTm (weakenCutoffTyVar C0 k45) (tsubstTm (weakenTrace d32 k45) S24 s16))))) (fun (x38 : (Index TmVar)) =>
      (fun (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 abs (let IHT := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT k45 d32 S24)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d32) k45 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TmVar k45) d32 S24) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 app (IHt125 k45 d32 S24) (IHt126 k45 d32 S24))) (fun (T : Ty) (t : Tm) IHt125 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tabs (let IHT := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT k45 d32 S24)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d32) k45 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt125 (HS TyVar k45) d32 S24) (f_equal2 tshiftTm eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt125 (T : Ty) (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 tapp (IHt125 k45 d32 S24) (let IHT := (tsubst_tshift0_Ty_comm_ind T) in
      (IHT k45 d32 S24)))) (fun (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal2 prod (IHt125 k45 d32 S24) (IHt126 k45 d32 S24))) (fun (p : Pat) (t1 : Tm) IHt125 (t2 : Tm) IHt126 (k45 : Hvl) (d32 : (Trace TyVar)) (S24 : Ty) =>
      (f_equal3 lett eq_refl (IHt125 k45 d32 S24) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TyVar d32) k45 (bindPat p)) eq_refl (f_equal2 tshiftTm (weakenCutoffTyVar_append C0 k45 (bindPat p)) eq_refl)) (eq_trans (IHt126 (appendHvl k45 (bindPat p)) d32 S24) (f_equal2 tshiftTm (eq_sym (weakenCutoffTyVar_append C0 k45 (bindPat p))) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d32 k45 (bindPat p))) eq_refl eq_refl))))))).
  End SubstShiftCommInd.
  Section SubstShiftComm.
    Definition tsubstTy_tshiftTy0_comm (S37 : Ty) : (forall (d44 : (Trace TyVar)) (S36 : Ty) ,
      ((tsubstTy (XS TyVar d44) S36 (tshiftTy C0 S37)) =
      (tshiftTy C0 (tsubstTy d44 S36 S37)))) := (tsubst_tshift0_Ty_comm_ind S37 H0).
    Definition substTm_shiftTm0_comm (s17 : Tm) : (forall (d44 : (Trace TmVar)) (s16 : Tm) ,
      ((substTm (XS TmVar d44) s16 (shiftTm C0 s17)) =
      (shiftTm C0 (substTm d44 s16 s17)))) := (subst_shift0_Tm_comm_ind s17 H0).
    Definition substTm_tshiftTm0_comm (s17 : Tm) : (forall (d44 : (Trace TmVar)) (s16 : Tm) ,
      ((substTm (XS TyVar d44) s16 (tshiftTm C0 s17)) =
      (tshiftTm C0 (substTm d44 s16 s17)))) := (subst_tshift0_Tm_comm_ind s17 H0).
    Definition tsubstTm_shiftTm0_comm (s16 : Tm) : (forall (d44 : (Trace TyVar)) (S36 : Ty) ,
      ((tsubstTm d44 S36 (shiftTm C0 s16)) =
      (shiftTm C0 (tsubstTm d44 S36 s16)))) := (tsubst_shift0_Tm_comm_ind s16 H0).
    Definition tsubstTm_tshiftTm0_comm (s16 : Tm) : (forall (d44 : (Trace TyVar)) (S36 : Ty) ,
      ((tsubstTm (XS TyVar d44) S36 (tshiftTm C0 s16)) =
      (tshiftTm C0 (tsubstTm d44 S36 s16)))) := (tsubst_tshift0_Tm_comm_ind s16 H0).
  End SubstShiftComm.
  Section SubstSubordInd.
    Definition tsubst_TmVar_Ty_ind := (ind_Ty (fun (S36 : Ty) =>
      (forall (k57 : Hvl) (d44 : (Trace TyVar)) (S37 : Ty) ,
        ((tsubstTy (weakenTrace (XS TmVar d44) k57) S37 S36) =
        (tsubstTy (weakenTrace d44 k57) S37 S36)))) (fun (X32 : (Index TyVar)) =>
      (fun (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
        (tsubstIndex_shiftTy0_comm_ind d44 S36 k57 X32))) (fun (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 tarrow (IHT6 k57 d44 S36) (IHT7 k57 d44 S36))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 tall (IHT6 k57 d44 S36) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar (XS TmVar d44) k57 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHT7 (HS TyVar k57) d44 S36) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar d44 k57 (HS TyVar H0))) eq_refl eq_refl))))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 tprod (IHT6 k57 d44 S36) (IHT7 k57 d44 S36)))).
    Definition tsubst_TmVar_Tm_ind := (ind_Tm (fun (s16 : Tm) =>
      (forall (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) ,
        ((tsubstTm (weakenTrace (XS TmVar d44) k57) S36 s16) =
        (tsubstTm (weakenTrace d44 k57) S36 s16)))) (fun (x41 : (Index TmVar)) =>
      (fun (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt134 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 abs (let IHT := (tsubst_TmVar_Ty_ind T) in
      (IHT k57 d44 S36)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d44) k57 (HS TmVar H0)) eq_refl eq_refl) (eq_trans (IHt134 (HS TmVar k57) d44 S36) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d44 k57 (HS TmVar H0))) eq_refl eq_refl))))) (fun (t1 : Tm) IHt134 (t2 : Tm) IHt135 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 app (IHt134 k57 d44 S36) (IHt135 k57 d44 S36))) (fun (T : Ty) (t : Tm) IHt134 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 tabs (let IHT := (tsubst_TmVar_Ty_ind T) in
      (IHT k57 d44 S36)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d44) k57 (HS TyVar H0)) eq_refl eq_refl) (eq_trans (IHt134 (HS TyVar k57) d44 S36) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d44 k57 (HS TyVar H0))) eq_refl eq_refl))))) (fun (t : Tm) IHt134 (T : Ty) (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 tapp (IHt134 k57 d44 S36) (let IHT := (tsubst_TmVar_Ty_ind T) in
      (IHT k57 d44 S36)))) (fun (t1 : Tm) IHt134 (t2 : Tm) IHt135 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal2 prod (IHt134 k57 d44 S36) (IHt135 k57 d44 S36))) (fun (p : Pat) (t1 : Tm) IHt134 (t2 : Tm) IHt135 (k57 : Hvl) (d44 : (Trace TyVar)) (S36 : Ty) =>
      (f_equal3 lett eq_refl (IHt134 k57 d44 S36) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar (XS TmVar d44) k57 (bindPat p)) eq_refl eq_refl) (eq_trans (IHt135 (appendHvl k57 (bindPat p)) d44 S36) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d44 k57 (bindPat p))) eq_refl eq_refl)))))).
  End SubstSubordInd.
  Section SubstSubord.
    Definition tsubstTy_TmVar (S37 : Ty) : (forall (d44 : (Trace TyVar)) (S36 : Ty) ,
      ((tsubstTy (XS TmVar d44) S36 S37) =
      (tsubstTy d44 S36 S37))) := (tsubst_TmVar_Ty_ind S37 H0).
    Definition tsubstTm_TmVar (s16 : Tm) : (forall (d44 : (Trace TyVar)) (S36 : Ty) ,
      ((tsubstTm (XS TmVar d44) S36 s16) =
      (tsubstTm d44 S36 s16))) := (tsubst_TmVar_Tm_ind s16 H0).
  End SubstSubord.
End ShiftSubstInteraction.
 Hint Rewrite substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection tsubstTy0_tshiftTy0_reflection : interaction.
 Hint Rewrite substTm0_shiftTm0_reflection tsubstTm0_tshiftTm0_reflection tsubstTy0_tshiftTy0_reflection : reflection.
 Hint Rewrite substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm tsubstTy_tshiftTy0_comm : interaction.
 Hint Rewrite substTm_shiftTm0_comm substTm_tshiftTm0_comm tsubstTm_shiftTm0_comm tsubstTm_tshiftTm0_comm tsubstTy_tshiftTy0_comm : subst_shift0.
 Hint Rewrite tsubstTm_TmVar tsubstTy_TmVar : interaction.
 Hint Rewrite tsubstTm_TmVar tsubstTy_TmVar : subst_shift0.
 Hint Rewrite shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm tshiftTy_tsubstTy0_comm : interaction.
 Hint Rewrite shiftTm_substTm0_comm shiftTm_tsubstTm0_comm tshiftTm_substTm0_comm tshiftTm_tsubstTm0_comm tshiftTy_tsubstTy0_comm : shift_subst0.
Section SubstSubstInteraction.
  Section SubstSubstIndexCommInd.
    Lemma substTm_substIndex0_commright_ind (d44 : (Trace TmVar)) (s16 : Tm) (s17 : Tm) :
      (forall (k57 : Hvl) (x41 : (Index TmVar)) ,
        ((substTm (weakenTrace d44 k57) s16 (substIndex (weakenTrace X0 k57) s17 x41)) =
        (substTm (weakenTrace X0 k57) (substTm d44 s16 s17) (substIndex (weakenTrace (XS TmVar d44) k57) s16 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTm_substIndex0_commright_ind (d44 : (Trace TyVar)) (S36 : Ty) (s16 : Tm) :
      (forall (k57 : Hvl) (x41 : (Index TmVar)) ,
        ((tsubstTm (weakenTrace d44 k57) S36 (substIndex (weakenTrace X0 k57) s16 x41)) =
        (substIndex (weakenTrace X0 k57) (tsubstTm d44 S36 s16) x41))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma tsubstTy_tsubstIndex0_commright_ind (d44 : (Trace TyVar)) (S36 : Ty) (S37 : Ty) :
      (forall (k57 : Hvl) (X32 : (Index TyVar)) ,
        ((tsubstTy (weakenTrace d44 k57) S36 (tsubstIndex (weakenTrace X0 k57) S37 X32)) =
        (tsubstTy (weakenTrace X0 k57) (tsubstTy d44 S36 S37) (tsubstIndex (weakenTrace (XS TyVar d44) k57) S36 X32)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
    Lemma substTy_tsubstIndex0_commleft_ind (d44 : (Trace TmVar)) (s16 : Tm) (S36 : Ty) :
      (forall (k57 : Hvl) (x41 : (Index TmVar)) ,
        ((substIndex (weakenTrace d44 k57) s16 x41) =
        (tsubstTm (weakenTrace X0 k57) S36 (substIndex (weakenTrace (XS TyVar d44) k57) s16 x41)))).
    Proof.
      needleGenericSubstSubstIndexCommInd .
    Qed.
  End SubstSubstIndexCommInd.
  Section SubstSubstCommInd.
    Definition tsubst_tsubst0_Ty_comm_ind := (ind_Ty (fun (S40 : Ty) =>
      (forall (k60 : Hvl) (d47 : (Trace TyVar)) (S41 : Ty) (S42 : Ty) ,
        ((tsubstTy (weakenTrace d47 k60) S41 (tsubstTy (weakenTrace X0 k60) S42 S40)) =
        (tsubstTy (weakenTrace X0 k60) (tsubstTy d47 S41 S42) (tsubstTy (weakenTrace (XS TyVar d47) k60) S41 S40))))) (fun (X37 : (Index TyVar)) =>
      (fun (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
        (tsubstTy_tsubstIndex0_commright_ind d47 S40 S41 k60 X37))) (fun (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      eq_refl) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 tarrow (IHT6 k60 d47 S40 S41) (IHT7 k60 d47 S40 S41))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 tall (IHT6 k60 d47 S40 S41) (eq_trans (f_equal3 tsubstTy (weakenTrace_append TyVar d47 k60 (HS TyVar H0)) eq_refl (f_equal3 tsubstTy (weakenTrace_append TyVar X0 k60 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHT7 (HS TyVar k60) d47 S40 S41) (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar X0 k60 (HS TyVar H0))) eq_refl (f_equal3 tsubstTy (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k60 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (T1 : Ty) IHT6 (T2 : Ty) IHT7 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 tprod (IHT6 k60 d47 S40 S41) (IHT7 k60 d47 S40 S41)))).
    Definition subst_subst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k60 : Hvl) (d47 : (Trace TmVar)) (s19 : Tm) (s20 : Tm) ,
        ((substTm (weakenTrace d47 k60) s19 (substTm (weakenTrace X0 k60) s20 s18)) =
        (substTm (weakenTrace X0 k60) (substTm d47 s19 s20) (substTm (weakenTrace (XS TmVar d47) k60) s19 s18))))) (fun (x47 : (Index TmVar)) =>
      (fun (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
        (substTm_substIndex0_commright_ind d47 s18 s19 k60 x47))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k60) d47 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k60 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 app (IHt152 k60 d47 s18 s19) (IHt153 k60 d47 s18 s19))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (HS TyVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k60) d47 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (HS TyVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k60 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 tapp (IHt152 k60 d47 s18 s19) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal2 prod (IHt152 k60 d47 s18 s19) (IHt153 k60 d47 s18 s19))) (fun (p : Pat) (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) =>
      (f_equal3 lett eq_refl (IHt152 k60 d47 s18 s19) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (bindPat p)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k60 (bindPat p)) d47 s18 s19) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (bindPat p))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TmVar d47) k60 (bindPat p))) eq_refl eq_refl))))))).
    Definition subst_tsubst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k60 : Hvl) (d47 : (Trace TmVar)) (s19 : Tm) (S40 : Ty) ,
        ((substTm (weakenTrace d47 k60) s19 (tsubstTm (weakenTrace X0 k60) S40 s18)) =
        (tsubstTm (weakenTrace X0 k60) S40 (substTm (weakenTrace (XS TyVar d47) k60) s19 s18))))) (fun (x47 : (Index TmVar)) =>
      (fun (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
        (substTy_tsubstIndex0_commleft_ind d47 s18 S40 k60 x47))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k60) d47 s18 S40) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (HS TmVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k60 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal2 app (IHt152 k60 d47 s18 S40) (IHt153 k60 d47 s18 S40))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (HS TyVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k60) d47 s18 S40) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (HS TyVar H0))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k60 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal2 tapp (IHt152 k60 d47 s18 S40) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal2 prod (IHt152 k60 d47 s18 S40) (IHt153 k60 d47 s18 S40))) (fun (p : Pat) (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) =>
      (f_equal3 lett eq_refl (IHt152 k60 d47 s18 S40) (eq_trans (f_equal3 substTm (weakenTrace_append TmVar d47 k60 (bindPat p)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k60 (bindPat p)) d47 s18 S40) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (bindPat p))) eq_refl (f_equal3 substTm (eq_sym (weakenTrace_append TmVar (XS TyVar d47) k60 (bindPat p))) eq_refl eq_refl))))))).
    Definition tsubst_subst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s19 : Tm) ,
        ((tsubstTm (weakenTrace d47 k60) S40 (substTm (weakenTrace X0 k60) s19 s18)) =
        (substTm (weakenTrace X0 k60) (tsubstTm d47 S40 s19) (tsubstTm (weakenTrace d47 k60) S40 s18))))) (fun (x47 : (Index TmVar)) =>
      (fun (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
        (tsubstTm_substIndex0_commright_ind d47 S40 s18 k60 x47))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal2 abs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (HS TmVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k60) d47 S40 s18) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d47 k60 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal2 app (IHt152 k60 d47 S40 s18) (IHt153 k60 d47 S40 s18))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal2 tabs eq_refl (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (HS TyVar H0)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k60) d47 S40 s18) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (HS TyVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d47 k60 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal2 tapp (IHt152 k60 d47 S40 s18) eq_refl)) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal2 prod (IHt152 k60 d47 S40 s18) (IHt153 k60 d47 S40 s18))) (fun (p : Pat) (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) =>
      (f_equal3 lett eq_refl (IHt152 k60 d47 S40 s18) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (bindPat p)) eq_refl (f_equal3 substTm (weakenTrace_append TmVar X0 k60 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k60 (bindPat p)) d47 S40 s18) (f_equal3 substTm (eq_sym (weakenTrace_append TmVar X0 k60 (bindPat p))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar d47 k60 (bindPat p))) eq_refl eq_refl))))))).
    Definition tsubst_tsubst0_Tm_comm_ind := (ind_Tm (fun (s18 : Tm) =>
      (forall (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
        ((tsubstTm (weakenTrace d47 k60) S40 (tsubstTm (weakenTrace X0 k60) S41 s18)) =
        (tsubstTm (weakenTrace X0 k60) (tsubstTy d47 S40 S41) (tsubstTm (weakenTrace (XS TyVar d47) k60) S40 s18))))) (fun (x47 : (Index TmVar)) =>
      (fun (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
        eq_refl)) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 abs (let IHT := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT k60 d47 S40 S41)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (HS TmVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (HS TmVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TmVar k60) d47 S40 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (HS TmVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k60 (HS TmVar H0))) eq_refl eq_refl)))))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 app (IHt152 k60 d47 S40 S41) (IHt153 k60 d47 S40 S41))) (fun (T : Ty) (t : Tm) IHt152 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 tabs (let IHT := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT k60 d47 S40 S41)) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (HS TyVar H0)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (HS TyVar H0)) eq_refl eq_refl)) (eq_trans (IHt152 (HS TyVar k60) d47 S40 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (HS TyVar H0))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k60 (HS TyVar H0))) eq_refl eq_refl)))))) (fun (t : Tm) IHt152 (T : Ty) (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 tapp (IHt152 k60 d47 S40 S41) (let IHT := (tsubst_tsubst0_Ty_comm_ind T) in
      (IHT k60 d47 S40 S41)))) (fun (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal2 prod (IHt152 k60 d47 S40 S41) (IHt153 k60 d47 S40 S41))) (fun (p : Pat) (t1 : Tm) IHt152 (t2 : Tm) IHt153 (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) =>
      (f_equal3 lett eq_refl (IHt152 k60 d47 S40 S41) (eq_trans (f_equal3 tsubstTm (weakenTrace_append TyVar d47 k60 (bindPat p)) eq_refl (f_equal3 tsubstTm (weakenTrace_append TyVar X0 k60 (bindPat p)) eq_refl eq_refl)) (eq_trans (IHt153 (appendHvl k60 (bindPat p)) d47 S40 S41) (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar X0 k60 (bindPat p))) eq_refl (f_equal3 tsubstTm (eq_sym (weakenTrace_append TyVar (XS TyVar d47) k60 (bindPat p))) eq_refl eq_refl))))))).
  End SubstSubstCommInd.
  Section SubstSubstComm.
    Definition tsubstTy_tsubstTy0_comm (S42 : Ty) : (forall (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
      ((tsubstTy d47 S40 (tsubstTy X0 S41 S42)) =
      (tsubstTy X0 (tsubstTy d47 S40 S41) (tsubstTy (XS TyVar d47) S40 S42)))) := (tsubst_tsubst0_Ty_comm_ind S42 H0).
    Definition substTm_substTm0_comm (s20 : Tm) : (forall (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) ,
      ((substTm d47 s18 (substTm X0 s19 s20)) =
      (substTm X0 (substTm d47 s18 s19) (substTm (XS TmVar d47) s18 s20)))) := (subst_subst0_Tm_comm_ind s20 H0).
    Definition substTm_tsubstTm0_comm (s19 : Tm) : (forall (d47 : (Trace TmVar)) (s18 : Tm) (S40 : Ty) ,
      ((substTm d47 s18 (tsubstTm X0 S40 s19)) =
      (tsubstTm X0 S40 (substTm (XS TyVar d47) s18 s19)))) := (subst_tsubst0_Tm_comm_ind s19 H0).
    Definition tsubstTm_substTm0_comm (s19 : Tm) : (forall (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) ,
      ((tsubstTm d47 S40 (substTm X0 s18 s19)) =
      (substTm X0 (tsubstTm d47 S40 s18) (tsubstTm d47 S40 s19)))) := (tsubst_subst0_Tm_comm_ind s19 H0).
    Definition tsubstTm_tsubstTm0_comm (s18 : Tm) : (forall (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
      ((tsubstTm d47 S40 (tsubstTm X0 S41 s18)) =
      (tsubstTm X0 (tsubstTy d47 S40 S41) (tsubstTm (XS TyVar d47) S40 s18)))) := (tsubst_tsubst0_Tm_comm_ind s18 H0).
  End SubstSubstComm.
  Section WeakenSubst.
    Lemma weakenTy_tsubstTy  :
      (forall (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (S41 : Ty) ,
        ((weakenTy (tsubstTy d47 S40 S41) k60) =
        (tsubstTy (weakenTrace d47 k60) S40 (weakenTy S41 k60)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_substTm  :
      (forall (k60 : Hvl) (d47 : (Trace TmVar)) (s18 : Tm) (s19 : Tm) ,
        ((weakenTm (substTm d47 s18 s19) k60) =
        (substTm (weakenTrace d47 k60) s18 (weakenTm s19 k60)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
    Lemma weakenTm_tsubstTm  :
      (forall (k60 : Hvl) (d47 : (Trace TyVar)) (S40 : Ty) (s18 : Tm) ,
        ((weakenTm (tsubstTm d47 S40 s18) k60) =
        (tsubstTm (weakenTrace d47 k60) S40 (weakenTm s18 k60)))).
    Proof.
      needleGenericWeakenSubst .
    Qed.
  End WeakenSubst.
  Section WeakenAppend.
    Lemma weakenTy_append  :
      (forall (S40 : Ty) (k60 : Hvl) (k61 : Hvl) ,
        ((weakenTy (weakenTy S40 k60) k61) =
        (weakenTy S40 (appendHvl k60 k61)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenPat_append  :
      (forall (p16 : Pat) (k60 : Hvl) (k61 : Hvl) ,
        ((weakenPat (weakenPat p16 k60) k61) =
        (weakenPat p16 (appendHvl k60 k61)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
    Lemma weakenTm_append  :
      (forall (s18 : Tm) (k60 : Hvl) (k61 : Hvl) ,
        ((weakenTm (weakenTm s18 k60) k61) =
        (weakenTm s18 (appendHvl k60 k61)))).
    Proof.
      needleGenericWeakenAppend .
    Qed.
  End WeakenAppend.
End SubstSubstInteraction.
 Hint Rewrite substTm_substTm0_comm tsubstTm_tsubstTm0_comm tsubstTy_tsubstTy0_comm : interaction.
 Hint Rewrite substTm_substTm0_comm tsubstTm_tsubstTm0_comm tsubstTy_tsubstTy0_comm : subst_subst0.
 Hint Rewrite weakenTm_shiftTm weakenTm_tshiftTm weakenTy_tshiftTy : weaken_shift.
 Hint Rewrite weakenTm_substTm weakenTm_tsubstTm weakenTy_tsubstTy : weaken_subst.
Section WellFormed.
  Fixpoint wfindex {a : Namespace} (k60 : Hvl) (i : (Index a)) {struct k60} : Prop :=
    match k60 with
      | (H0) => False
      | (HS b k60) => match (eq_namespace_dec a b) with
        | (left e) => match i with
          | (I0) => True
          | (IS i) => (wfindex k60 i)
        end
        | (right n) => (wfindex k60 i)
      end
    end.
  Inductive wfTy (k60 : Hvl) : Ty -> Prop :=
    | wfTy_tvar
        {X43 : (Index TyVar)}
        (wfi : (wfindex k60 X43)) :
        (wfTy k60 (tvar X43))
    | wfTy_top : (wfTy k60 (top))
    | wfTy_tarrow {T98 : Ty}
        (wf : (wfTy k60 T98)) {T99 : Ty}
        (wf0 : (wfTy k60 T99)) :
        (wfTy k60 (tarrow T98 T99))
    | wfTy_tall {T100 : Ty}
        (wf : (wfTy k60 T100))
        {T101 : Ty}
        (wf0 : (wfTy (HS TyVar k60) T101))
        : (wfTy k60 (tall T100 T101))
    | wfTy_tprod {T102 : Ty}
        (wf : (wfTy k60 T102))
        {T103 : Ty}
        (wf0 : (wfTy k60 T103)) :
        (wfTy k60 (tprod T102 T103)).
  Inductive wfPat (k60 : Hvl) : Pat -> Prop :=
    | wfPat_pvar :
        (wfPat k60 (pvar))
    | wfPat_pprod {p18 : Pat}
        (wf : (wfPat k60 p18))
        {p19 : Pat}
        (wf0 : (wfPat k60 p19)) :
        (wfPat k60 (pprod p18 p19)).
  Inductive wfTm (k60 : Hvl) : Tm -> Prop :=
    | wfTm_var
        {x52 : (Index TmVar)}
        (wfi : (wfindex k60 x52)) :
        (wfTm k60 (var x52))
    | wfTm_abs {T98 : Ty}
        (wf : (wfTy k60 T98))
        {t170 : Tm}
        (wf0 : (wfTm (HS TmVar k60) t170))
        : (wfTm k60 (abs T98 t170))
    | wfTm_app {t171 : Tm}
        (wf : (wfTm k60 t171))
        {t172 : Tm}
        (wf0 : (wfTm k60 t172)) :
        (wfTm k60 (app t171 t172))
    | wfTm_tabs {T99 : Ty}
        (wf : (wfTy k60 T99))
        {t173 : Tm}
        (wf0 : (wfTm (HS TyVar k60) t173))
        : (wfTm k60 (tabs T99 t173))
    | wfTm_tapp {t174 : Tm}
        (wf : (wfTm k60 t174))
        {T100 : Ty}
        (wf0 : (wfTy k60 T100)) :
        (wfTm k60 (tapp t174 T100))
    | wfTm_prod {t175 : Tm}
        (wf : (wfTm k60 t175))
        {t176 : Tm}
        (wf0 : (wfTm k60 t176)) :
        (wfTm k60 (prod t175 t176))
    | wfTm_lett {p18 : Pat}
        (wf : (wfPat k60 p18))
        {t177 : Tm}
        (wf0 : (wfTm k60 t177))
        {t178 : Tm}
        (wf1 : (wfTm (appendHvl k60 (bindPat p18)) t178))
        :
        (wfTm k60 (lett p18 t177 t178)).
  Scheme ind_wfTy := Induction for wfTy Sort Prop.
  Scheme ind_wfPat := Induction for wfPat Sort Prop.
  Scheme ind_wfTm := Induction for wfTm Sort Prop.
End WellFormed.
Section ShiftWellFormed.
  Inductive shifthvl_TmVar : (forall (c12 : (Cutoff TmVar)) (k60 : Hvl) (k61 : Hvl) ,
    Prop) :=
    | shifthvl_TmVar_here
        (k60 : Hvl) :
        (shifthvl_TmVar C0 k60 (HS TmVar k60))
    | shifthvl_TmVar_there_TmVar
        (c12 : (Cutoff TmVar))
        (k60 : Hvl) (k61 : Hvl) :
        (shifthvl_TmVar c12 k60 k61) -> (shifthvl_TmVar (CS c12) (HS TmVar k60) (HS TmVar k61))
    | shifthvl_TmVar_there_TyVar
        (c12 : (Cutoff TmVar))
        (k60 : Hvl) (k61 : Hvl) :
        (shifthvl_TmVar c12 k60 k61) -> (shifthvl_TmVar c12 (HS TyVar k60) (HS TyVar k61)).
  Inductive shifthvl_TyVar : (forall (c12 : (Cutoff TyVar)) (k60 : Hvl) (k61 : Hvl) ,
    Prop) :=
    | shifthvl_TyVar_here
        (k60 : Hvl) :
        (shifthvl_TyVar C0 k60 (HS TyVar k60))
    | shifthvl_TyVar_there_TmVar
        (c12 : (Cutoff TyVar))
        (k60 : Hvl) (k61 : Hvl) :
        (shifthvl_TyVar c12 k60 k61) -> (shifthvl_TyVar c12 (HS TmVar k60) (HS TmVar k61))
    | shifthvl_TyVar_there_TyVar
        (c12 : (Cutoff TyVar))
        (k60 : Hvl) (k61 : Hvl) :
        (shifthvl_TyVar c12 k60 k61) -> (shifthvl_TyVar (CS c12) (HS TyVar k60) (HS TyVar k61)).
  Lemma weaken_shifthvl_TmVar  :
    (forall (k60 : Hvl) {c12 : (Cutoff TmVar)} {k61 : Hvl} {k62 : Hvl} ,
      (shifthvl_TmVar c12 k61 k62) -> (shifthvl_TmVar (weakenCutoffTmVar c12 k60) (appendHvl k61 k60) (appendHvl k62 k60))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma weaken_shifthvl_TyVar  :
    (forall (k60 : Hvl) {c12 : (Cutoff TyVar)} {k61 : Hvl} {k62 : Hvl} ,
      (shifthvl_TyVar c12 k61 k62) -> (shifthvl_TyVar (weakenCutoffTyVar c12 k60) (appendHvl k61 k60) (appendHvl k62 k60))).
  Proof.
    needleGenericWeakenHVarlistInsert .
  Qed.
  Lemma shift_wfindex_TmVar  :
    (forall (c12 : (Cutoff TmVar)) (k60 : Hvl) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) (x52 : (Index TmVar)) ,
      (wfindex k60 x52) -> (wfindex k61 (shiftIndex c12 x52))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma shift_wfindex_TyVar  :
    (forall (c12 : (Cutoff TmVar)) (k60 : Hvl) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) (X43 : (Index TyVar)) ,
      (wfindex k60 X43) -> (wfindex k61 X43)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TmVar  :
    (forall (c12 : (Cutoff TyVar)) (k60 : Hvl) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) (x52 : (Index TmVar)) ,
      (wfindex k60 x52) -> (wfindex k61 x52)).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Lemma tshift_wfindex_TyVar  :
    (forall (c12 : (Cutoff TyVar)) (k60 : Hvl) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) (X43 : (Index TyVar)) ,
      (wfindex k60 X43) -> (wfindex k61 (tshiftIndex c12 X43))).
  Proof.
    needleGenericShiftWellFormedIndex .
  Qed.
  Definition shift_wfTy : (forall (k60 : Hvl) ,
    (forall (S40 : Ty) (wf : (wfTy k60 S40)) ,
      (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
        (shifthvl_TmVar c12 k60 k61) -> (wfTy k61 S40)))) := (ind_wfTy (fun (k60 : Hvl) (S40 : Ty) (wf : (wfTy k60 S40)) =>
    (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
      (shifthvl_TmVar c12 k60 k61) -> (wfTy k61 S40))) (fun (k60 : Hvl) (X43 : (Index TyVar)) (wfi : (wfindex k60 X43)) (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTy_tvar k61 (shift_wfindex_TyVar c12 k60 k61 ins X43 wfi))) (fun (k60 : Hvl) (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTy_top k61)) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy k60 T2)) IHT7 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTy_tarrow k61 (IHT6 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHT7 c12 k61 (weaken_shifthvl_TmVar H0 ins)))) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy (HS TyVar k60) T2)) IHT7 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTy_tall k61 (IHT6 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHT7 c12 (HS TyVar k61) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy k60 T2)) IHT7 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTy_tprod k61 (IHT6 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHT7 c12 k61 (weaken_shifthvl_TmVar H0 ins))))).
  Definition tshift_wfTy : (forall (k60 : Hvl) ,
    (forall (S40 : Ty) (wf : (wfTy k60 S40)) ,
      (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
        (shifthvl_TyVar c12 k60 k61) -> (wfTy k61 (tshiftTy c12 S40))))) := (ind_wfTy (fun (k60 : Hvl) (S40 : Ty) (wf : (wfTy k60 S40)) =>
    (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
      (shifthvl_TyVar c12 k60 k61) -> (wfTy k61 (tshiftTy c12 S40)))) (fun (k60 : Hvl) (X43 : (Index TyVar)) (wfi : (wfindex k60 X43)) (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTy_tvar k61 (tshift_wfindex_TyVar c12 k60 k61 ins X43 wfi))) (fun (k60 : Hvl) (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTy_top k61)) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy k60 T2)) IHT7 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTy_tarrow k61 (IHT6 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHT7 c12 k61 (weaken_shifthvl_TyVar H0 ins)))) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy (HS TyVar k60) T2)) IHT7 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTy_tall k61 (IHT6 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHT7 (CS c12) (HS TyVar k61) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k60 : Hvl) (T1 : Ty) (wf : (wfTy k60 T1)) IHT6 (T2 : Ty) (wf0 : (wfTy k60 T2)) IHT7 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTy_tprod k61 (IHT6 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHT7 c12 k61 (weaken_shifthvl_TyVar H0 ins))))).
  Definition shift_wfPat : (forall (k60 : Hvl) ,
    (forall (p18 : Pat) (wf : (wfPat k60 p18)) ,
      (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
        (shifthvl_TmVar c12 k60 k61) -> (wfPat k61 p18)))) := (fun (k60 : Hvl) =>
    (ind_wfPat k60 (fun (p18 : Pat) (wf : (wfPat k60 p18)) =>
      (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
        (shifthvl_TmVar c12 k60 k61) -> (wfPat k61 p18))) (fun (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
      (wfPat_pvar k61)) (fun (p1 : Pat) (wf : (wfPat k60 p1)) IHp1 (p2 : Pat) (wf0 : (wfPat k60 p2)) IHp2 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
      (wfPat_pprod k61 (IHp1 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHp2 c12 k61 (weaken_shifthvl_TmVar H0 ins)))))).
  Definition tshift_wfPat : (forall (k60 : Hvl) ,
    (forall (p18 : Pat) (wf : (wfPat k60 p18)) ,
      (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
        (shifthvl_TyVar c12 k60 k61) -> (wfPat k61 p18)))) := (fun (k60 : Hvl) =>
    (ind_wfPat k60 (fun (p18 : Pat) (wf : (wfPat k60 p18)) =>
      (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
        (shifthvl_TyVar c12 k60 k61) -> (wfPat k61 p18))) (fun (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
      (wfPat_pvar k61)) (fun (p1 : Pat) (wf : (wfPat k60 p1)) IHp1 (p2 : Pat) (wf0 : (wfPat k60 p2)) IHp2 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
      (wfPat_pprod k61 (IHp1 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHp2 c12 k61 (weaken_shifthvl_TyVar H0 ins)))))).
  Definition shift_wfTm : (forall (k60 : Hvl) ,
    (forall (s18 : Tm) (wf : (wfTm k60 s18)) ,
      (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
        (shifthvl_TmVar c12 k60 k61) -> (wfTm k61 (shiftTm c12 s18))))) := (ind_wfTm (fun (k60 : Hvl) (s18 : Tm) (wf : (wfTm k60 s18)) =>
    (forall (c12 : (Cutoff TmVar)) (k61 : Hvl) ,
      (shifthvl_TmVar c12 k60 k61) -> (wfTm k61 (shiftTm c12 s18)))) (fun (k60 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k60 x52)) (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_var k61 (shift_wfindex_TmVar c12 k60 k61 ins x52 wfi))) (fun (k60 : Hvl) (T : Ty) (wf : (wfTy k60 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k60) t)) IHt170 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_abs k61 (shift_wfTy k60 T wf c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt170 (CS c12) (HS TmVar k61) (weaken_shifthvl_TmVar (HS TmVar H0) ins)))) (fun (k60 : Hvl) (t1 : Tm) (wf : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k60 t2)) IHt171 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_app k61 (IHt170 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt171 c12 k61 (weaken_shifthvl_TmVar H0 ins)))) (fun (k60 : Hvl) (T : Ty) (wf : (wfTy k60 T)) (t : Tm) (wf0 : (wfTm (HS TyVar k60) t)) IHt170 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_tabs k61 (shift_wfTy k60 T wf c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt170 c12 (HS TyVar k61) (weaken_shifthvl_TmVar (HS TyVar H0) ins)))) (fun (k60 : Hvl) (t : Tm) (wf : (wfTm k60 t)) IHt170 (T : Ty) (wf0 : (wfTy k60 T)) (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_tapp k61 (IHt170 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (shift_wfTy k60 T wf0 c12 k61 (weaken_shifthvl_TmVar H0 ins)))) (fun (k60 : Hvl) (t1 : Tm) (wf : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k60 t2)) IHt171 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_prod k61 (IHt170 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt171 c12 k61 (weaken_shifthvl_TmVar H0 ins)))) (fun (k60 : Hvl) (p : Pat) (wf : (wfPat k60 p)) (t1 : Tm) (wf0 : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm (appendHvl k60 (bindPat p)) t2)) IHt171 (c12 : (Cutoff TmVar)) (k61 : Hvl) (ins : (shifthvl_TmVar c12 k60 k61)) =>
    (wfTm_lett k61 (shift_wfPat k60 p wf c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt170 c12 k61 (weaken_shifthvl_TmVar H0 ins)) (IHt171 (weakenCutoffTmVar c12 (bindPat p)) (appendHvl k61 (bindPat p)) (weaken_shifthvl_TmVar (bindPat p) ins))))).
  Definition tshift_wfTm : (forall (k60 : Hvl) ,
    (forall (s18 : Tm) (wf : (wfTm k60 s18)) ,
      (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
        (shifthvl_TyVar c12 k60 k61) -> (wfTm k61 (tshiftTm c12 s18))))) := (ind_wfTm (fun (k60 : Hvl) (s18 : Tm) (wf : (wfTm k60 s18)) =>
    (forall (c12 : (Cutoff TyVar)) (k61 : Hvl) ,
      (shifthvl_TyVar c12 k60 k61) -> (wfTm k61 (tshiftTm c12 s18)))) (fun (k60 : Hvl) (x52 : (Index TmVar)) (wfi : (wfindex k60 x52)) (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_var k61 (tshift_wfindex_TmVar c12 k60 k61 ins x52 wfi))) (fun (k60 : Hvl) (T : Ty) (wf : (wfTy k60 T)) (t : Tm) (wf0 : (wfTm (HS TmVar k60) t)) IHt170 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_abs k61 (tshift_wfTy k60 T wf c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt170 c12 (HS TmVar k61) (weaken_shifthvl_TyVar (HS TmVar H0) ins)))) (fun (k60 : Hvl) (t1 : Tm) (wf : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k60 t2)) IHt171 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_app k61 (IHt170 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt171 c12 k61 (weaken_shifthvl_TyVar H0 ins)))) (fun (k60 : Hvl) (T : Ty) (wf : (wfTy k60 T)) (t : Tm) (wf0 : (wfTm (HS TyVar k60) t)) IHt170 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_tabs k61 (tshift_wfTy k60 T wf c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt170 (CS c12) (HS TyVar k61) (weaken_shifthvl_TyVar (HS TyVar H0) ins)))) (fun (k60 : Hvl) (t : Tm) (wf : (wfTm k60 t)) IHt170 (T : Ty) (wf0 : (wfTy k60 T)) (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_tapp k61 (IHt170 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (tshift_wfTy k60 T wf0 c12 k61 (weaken_shifthvl_TyVar H0 ins)))) (fun (k60 : Hvl) (t1 : Tm) (wf : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf0 : (wfTm k60 t2)) IHt171 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_prod k61 (IHt170 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt171 c12 k61 (weaken_shifthvl_TyVar H0 ins)))) (fun (k60 : Hvl) (p : Pat) (wf : (wfPat k60 p)) (t1 : Tm) (wf0 : (wfTm k60 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm (appendHvl k60 (bindPat p)) t2)) IHt171 (c12 : (Cutoff TyVar)) (k61 : Hvl) (ins : (shifthvl_TyVar c12 k60 k61)) =>
    (wfTm_lett k61 (tshift_wfPat k60 p wf c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt170 c12 k61 (weaken_shifthvl_TyVar H0 ins)) (IHt171 (weakenCutoffTyVar c12 (bindPat p)) (appendHvl k61 (bindPat p)) (weaken_shifthvl_TyVar (bindPat p) ins))))).
End ShiftWellFormed.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : infra.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : shift_wf.
 Hint Resolve shift_wfindex_TmVar shift_wfindex_TyVar tshift_wfindex_TmVar tshift_wfindex_TyVar : wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : infra.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : shift_wf.
 Hint Resolve shift_wfPat shift_wfTm shift_wfTy tshift_wfPat tshift_wfTm tshift_wfTy : wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : infra.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : shift_wf.
 Hint Constructors shifthvl_TmVar shifthvl_TyVar : wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : infra.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : shift_wf.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : weaken.
 Hint Resolve weaken_shifthvl_TmVar weaken_shifthvl_TyVar : wf.
Section SubstWellFormed.
  Inductive substhvl_TmVar (k60 : Hvl) : (forall (d47 : (Trace TmVar)) (k61 : Hvl) (k62 : Hvl) ,
    Prop) :=
    | substhvl_TmVar_here :
        (substhvl_TmVar k60 X0 (HS TmVar k60) k60)
    | substhvl_TmVar_there_TmVar
        {d47 : (Trace TmVar)}
        {k61 : Hvl} {k62 : Hvl} :
        (substhvl_TmVar k60 d47 k61 k62) -> (substhvl_TmVar k60 (XS TmVar d47) (HS TmVar k61) (HS TmVar k62))
    | substhvl_TmVar_there_TyVar
        {d47 : (Trace TmVar)}
        {k61 : Hvl} {k62 : Hvl} :
        (substhvl_TmVar k60 d47 k61 k62) -> (substhvl_TmVar k60 (XS TyVar d47) (HS TyVar k61) (HS TyVar k62)).
  Inductive substhvl_TyVar (k60 : Hvl) : (forall (d47 : (Trace TyVar)) (k61 : Hvl) (k62 : Hvl) ,
    Prop) :=
    | substhvl_TyVar_here :
        (substhvl_TyVar k60 X0 (HS TyVar k60) k60)
    | substhvl_TyVar_there_TmVar
        {d47 : (Trace TyVar)}
        {k61 : Hvl} {k62 : Hvl} :
        (substhvl_TyVar k60 d47 k61 k62) -> (substhvl_TyVar k60 (XS TmVar d47) (HS TmVar k61) (HS TmVar k62))
    | substhvl_TyVar_there_TyVar
        {d47 : (Trace TyVar)}
        {k61 : Hvl} {k62 : Hvl} :
        (substhvl_TyVar k60 d47 k61 k62) -> (substhvl_TyVar k60 (XS TyVar d47) (HS TyVar k61) (HS TyVar k62)).
  Lemma weaken_substhvl_TmVar  :
    (forall {k61 : Hvl} (k60 : Hvl) {d47 : (Trace TmVar)} {k62 : Hvl} {k63 : Hvl} ,
      (substhvl_TmVar k61 d47 k62 k63) -> (substhvl_TmVar k61 (weakenTrace d47 k60) (appendHvl k62 k60) (appendHvl k63 k60))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma weaken_substhvl_TyVar  :
    (forall {k61 : Hvl} (k60 : Hvl) {d47 : (Trace TyVar)} {k62 : Hvl} {k63 : Hvl} ,
      (substhvl_TyVar k61 d47 k62 k63) -> (substhvl_TyVar k61 (weakenTrace d47 k60) (appendHvl k62 k60) (appendHvl k63 k60))).
  Proof.
    needleGenericWeakenSubstHvl .
  Qed.
  Lemma substhvl_TmVar_wfindex_TmVar {k64 : Hvl} {s18 : Tm} (wft : (wfTm k64 s18)) :
    (forall {d47 : (Trace TmVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TmVar k64 d47 k65 k66) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k65 x52) -> (wfTm k66 (substIndex d47 s18 x52)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TyVar_wfindex_TyVar {k64 : Hvl} {S41 : Ty} (wft : (wfTy k64 S41)) :
    (forall {d47 : (Trace TyVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TyVar k64 d47 k65 k66) -> (forall {X43 : (Index TyVar)} ,
        (wfindex k65 X43) -> (wfTy k66 (tsubstIndex d47 S41 X43)))).
  Proof.
    needleGenericSubstHvlWfIndexHom .
  Qed.
  Lemma substhvl_TmVar_wfindex_TyVar {k64 : Hvl} :
    (forall {d47 : (Trace TmVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TmVar k64 d47 k65 k66) -> (forall {X43 : (Index TyVar)} ,
        (wfindex k65 X43) -> (wfindex k66 X43))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Lemma substhvl_TyVar_wfindex_TmVar {k64 : Hvl} :
    (forall {d47 : (Trace TyVar)} {k65 : Hvl} {k66 : Hvl} ,
      (substhvl_TyVar k64 d47 k65 k66) -> (forall {x52 : (Index TmVar)} ,
        (wfindex k65 x52) -> (wfindex k66 x52))).
  Proof.
    needleGenericSubstHvlWfIndexHet .
  Qed.
  Definition substhvl_TmVar_wfTy {k64 : Hvl} : (forall (k65 : Hvl) ,
    (forall (S41 : Ty) (wf0 : (wfTy k65 S41)) ,
      (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d47 k65 k66) -> (wfTy k66 S41)))) := (ind_wfTy (fun (k65 : Hvl) (S41 : Ty) (wf0 : (wfTy k65 S41)) =>
    (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
      (substhvl_TmVar k64 d47 k65 k66) -> (wfTy k66 S41))) (fun (k65 : Hvl) {X43 : (Index TyVar)} (wfi : (wfindex k65 X43)) {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTy_tvar k66 (substhvl_TmVar_wfindex_TyVar del wfi))) (fun (k65 : Hvl) {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTy_top k66)) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT7 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTy_tarrow k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHT7 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy (HS TyVar k65) T2)) IHT7 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTy_tall k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHT7 (weakenTrace d47 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT7 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTy_tprod k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHT7 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del))))).
  Definition substhvl_TyVar_wfTy {k64 : Hvl} {S41 : Ty} (wf : (wfTy k64 S41)) : (forall (k65 : Hvl) ,
    (forall (S42 : Ty) (wf0 : (wfTy k65 S42)) ,
      (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d47 k65 k66) -> (wfTy k66 (tsubstTy d47 S41 S42))))) := (ind_wfTy (fun (k65 : Hvl) (S42 : Ty) (wf0 : (wfTy k65 S42)) =>
    (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
      (substhvl_TyVar k64 d47 k65 k66) -> (wfTy k66 (tsubstTy d47 S41 S42)))) (fun (k65 : Hvl) {X43 : (Index TyVar)} (wfi : (wfindex k65 X43)) {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (substhvl_TyVar_wfindex_TyVar wf del wfi)) (fun (k65 : Hvl) {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTy_top k66)) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT7 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTy_tarrow k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHT7 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy (HS TyVar k65) T2)) IHT7 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTy_tall k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHT7 (weakenTrace d47 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (T1 : Ty) (wf0 : (wfTy k65 T1)) IHT6 (T2 : Ty) (wf1 : (wfTy k65 T2)) IHT7 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTy_tprod k66 (IHT6 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHT7 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del))))).
  Definition substhvl_TmVar_wfPat {k64 : Hvl} : (forall (k65 : Hvl) ,
    (forall (p19 : Pat) (wf0 : (wfPat k65 p19)) ,
      (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d47 k65 k66) -> (wfPat k66 p19)))) := (fun (k65 : Hvl) =>
    (ind_wfPat k65 (fun (p19 : Pat) (wf0 : (wfPat k65 p19)) =>
      (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d47 k65 k66) -> (wfPat k66 p19))) (fun {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
      (wfPat_pvar k66)) (fun (p1 : Pat) (wf0 : (wfPat k65 p1)) IHp1 (p2 : Pat) (wf1 : (wfPat k65 p2)) IHp2 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
      (wfPat_pprod k66 (IHp1 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHp2 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)))))).
  Definition substhvl_TyVar_wfPat {k64 : Hvl} : (forall (k65 : Hvl) ,
    (forall (p19 : Pat) (wf0 : (wfPat k65 p19)) ,
      (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d47 k65 k66) -> (wfPat k66 p19)))) := (fun (k65 : Hvl) =>
    (ind_wfPat k65 (fun (p19 : Pat) (wf0 : (wfPat k65 p19)) =>
      (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d47 k65 k66) -> (wfPat k66 p19))) (fun {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
      (wfPat_pvar k66)) (fun (p1 : Pat) (wf0 : (wfPat k65 p1)) IHp1 (p2 : Pat) (wf1 : (wfPat k65 p2)) IHp2 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
      (wfPat_pprod k66 (IHp1 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHp2 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)))))).
  Definition substhvl_TmVar_wfTm {k64 : Hvl} {s18 : Tm} (wf : (wfTm k64 s18)) : (forall (k65 : Hvl) ,
    (forall (s19 : Tm) (wf0 : (wfTm k65 s19)) ,
      (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
        (substhvl_TmVar k64 d47 k65 k66) -> (wfTm k66 (substTm d47 s18 s19))))) := (ind_wfTm (fun (k65 : Hvl) (s19 : Tm) (wf0 : (wfTm k65 s19)) =>
    (forall {d47 : (Trace TmVar)} {k66 : Hvl} ,
      (substhvl_TmVar k64 d47 k65 k66) -> (wfTm k66 (substTm d47 s18 s19)))) (fun (k65 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k65 x52)) {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (substhvl_TmVar_wfindex_TmVar wf del wfi)) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k65) t)) IHt170 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_abs k66 (substhvl_TmVar_wfTy k65 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt170 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k66) (weaken_substhvl_TmVar (HS TmVar H0) del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_app k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TyVar k65) t)) IHt170 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_tabs k66 (substhvl_TmVar_wfTy k65 T wf0 (weaken_substhvl_TmVar H0 del)) (IHt170 (weakenTrace d47 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TmVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm k65 t)) IHt170 (T : Ty) (wf1 : (wfTy k65 T)) {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_tapp k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (substhvl_TmVar_wfTy k65 T wf1 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_prod k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)))) (fun (k65 : Hvl) (p : Pat) (wf0 : (wfPat k65 p)) (t1 : Tm) (wf1 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf2 : (wfTm (appendHvl k65 (bindPat p)) t2)) IHt171 {d47 : (Trace TmVar)} {k66 : Hvl} (del : (substhvl_TmVar k64 d47 k65 k66)) =>
    (wfTm_lett k66 (substhvl_TmVar_wfPat k65 p wf0 (weaken_substhvl_TmVar H0 del)) (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TmVar H0 del)) (IHt171 (weakenTrace d47 (bindPat p)) (appendHvl k66 (bindPat p)) (weaken_substhvl_TmVar (bindPat p) del))))).
  Definition substhvl_TyVar_wfTm {k64 : Hvl} {S41 : Ty} (wf : (wfTy k64 S41)) : (forall (k65 : Hvl) ,
    (forall (s18 : Tm) (wf0 : (wfTm k65 s18)) ,
      (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
        (substhvl_TyVar k64 d47 k65 k66) -> (wfTm k66 (tsubstTm d47 S41 s18))))) := (ind_wfTm (fun (k65 : Hvl) (s18 : Tm) (wf0 : (wfTm k65 s18)) =>
    (forall {d47 : (Trace TyVar)} {k66 : Hvl} ,
      (substhvl_TyVar k64 d47 k65 k66) -> (wfTm k66 (tsubstTm d47 S41 s18)))) (fun (k65 : Hvl) {x52 : (Index TmVar)} (wfi : (wfindex k65 x52)) {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_var k66 (substhvl_TyVar_wfindex_TmVar del wfi))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TmVar k65) t)) IHt170 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_abs k66 (substhvl_TyVar_wfTy wf k65 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt170 (weakenTrace d47 (HS TmVar H0)) (HS TmVar k66) (weaken_substhvl_TyVar (HS TmVar H0) del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_app k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHt171 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (T : Ty) (wf0 : (wfTy k65 T)) (t : Tm) (wf1 : (wfTm (HS TyVar k65) t)) IHt170 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_tabs k66 (substhvl_TyVar_wfTy wf k65 T wf0 (weaken_substhvl_TyVar H0 del)) (IHt170 (weakenTrace d47 (HS TyVar H0)) (HS TyVar k66) (weaken_substhvl_TyVar (HS TyVar H0) del)))) (fun (k65 : Hvl) (t : Tm) (wf0 : (wfTm k65 t)) IHt170 (T : Ty) (wf1 : (wfTy k65 T)) {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_tapp k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (substhvl_TyVar_wfTy wf k65 T wf1 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (t1 : Tm) (wf0 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf1 : (wfTm k65 t2)) IHt171 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_prod k66 (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHt171 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)))) (fun (k65 : Hvl) (p : Pat) (wf0 : (wfPat k65 p)) (t1 : Tm) (wf1 : (wfTm k65 t1)) IHt170 (t2 : Tm) (wf2 : (wfTm (appendHvl k65 (bindPat p)) t2)) IHt171 {d47 : (Trace TyVar)} {k66 : Hvl} (del : (substhvl_TyVar k64 d47 k65 k66)) =>
    (wfTm_lett k66 (substhvl_TyVar_wfPat k65 p wf0 (weaken_substhvl_TyVar H0 del)) (IHt170 (weakenTrace d47 H0) k66 (weaken_substhvl_TyVar H0 del)) (IHt171 (weakenTrace d47 (bindPat p)) (appendHvl k66 (bindPat p)) (weaken_substhvl_TyVar (bindPat p) del))))).
End SubstWellFormed.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : infra.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : subst_wf.
 Hint Resolve substhvl_TmVar_wfindex_TmVar substhvl_TmVar_wfindex_TyVar substhvl_TyVar_wfindex_TmVar substhvl_TyVar_wfindex_TyVar : wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : infra.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : subst_wf.
 Hint Resolve substhvl_TmVar_wfPat substhvl_TmVar_wfTm substhvl_TmVar_wfTy substhvl_TyVar_wfPat substhvl_TyVar_wfTm substhvl_TyVar_wfTy : wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : infra.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst.
 Hint Constructors substhvl_TmVar substhvl_TyVar : subst_wf.
 Hint Constructors substhvl_TmVar substhvl_TyVar : wf.
Fixpoint subhvl_TmVar (k64 : Hvl) {struct k64} : Prop :=
  match k64 with
    | (H0) => True
    | (HS a k64) => match a with
      | (TmVar) => (subhvl_TmVar k64)
      | _ => False
    end
  end.
Lemma subhvl_TmVar_append  :
  (forall (k64 : Hvl) (k65 : Hvl) ,
    (subhvl_TmVar k64) -> (subhvl_TmVar k65) -> (subhvl_TmVar (appendHvl k64 k65))).
Proof.
  needleGenericSubHvlAppend .
Qed.
 Hint Resolve subhvl_TmVar_append : infra.
 Hint Resolve subhvl_TmVar_append : wf.
Lemma wfPat_strengthen_subhvl_TmVar  :
  (forall (k61 : Hvl) (k60 : Hvl) (p18 : Pat) ,
    (subhvl_TmVar k61) -> (wfPat (appendHvl k60 k61) (weakenPat p18 k61)) -> (wfPat k60 p18)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
Lemma wfTy_strengthen_subhvl_TmVar  :
  (forall (k63 : Hvl) (k62 : Hvl) (S40 : Ty) ,
    (subhvl_TmVar k63) -> (wfTy (appendHvl k62 k63) (weakenTy S40 k63)) -> (wfTy k62 S40)).
Proof.
  needleGenericWellformedStrengthen .
Qed.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat (appendHvl _ _) (weakenPat _ _)) |- _ => apply (wfPat_strengthen_subhvl_TmVar) in H
end : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H0 : (wfTy (appendHvl _ _) (weakenTy _ _)) |- _ => apply (wfTy_strengthen_subhvl_TmVar) in H0
end : infra wf.
Section Context.
  Inductive Env : Type :=
    | empty 
    | evar (G : Env) (T : Ty)
    | ebound (G : Env) (T : Ty).
  Fixpoint appendEnv (G : Env) (G0 : Env) : Env :=
    match G0 with
      | (empty) => G
      | (evar G1 T) => (evar (appendEnv G G1) T)
      | (ebound G1 T) => (ebound (appendEnv G G1) T)
    end.
  Fixpoint domainEnv (G : Env) : Hvl :=
    match G with
      | (empty) => H0
      | (evar G0 T) => (HS TmVar (domainEnv G0))
      | (ebound G0 T) => (HS TyVar (domainEnv G0))
    end.
  Lemma appendEnv_assoc  :
    (forall (G : Env) (G0 : Env) (G1 : Env) ,
      ((appendEnv G (appendEnv G0 G1)) =
      (appendEnv (appendEnv G G0) G1))).
  Proof.
    needleGenericAppendEnvAssoc .
  Qed.
  Lemma domainEnv_appendEnv  :
    (forall (G : Env) (G0 : Env) ,
      ((domainEnv (appendEnv G G0)) =
      (appendHvl (domainEnv G) (domainEnv G0)))).
  Proof.
    needleGenericDomainEnvAppendEnv .
  Qed.
  Fixpoint shiftEnv (c12 : (Cutoff TmVar)) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (shiftEnv c12 G0) T)
      | (ebound G0 T) => (ebound (shiftEnv c12 G0) T)
    end.
  Fixpoint tshiftEnv (c12 : (Cutoff TyVar)) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tshiftEnv c12 G0) (tshiftTy (weakenCutoffTyVar c12 (domainEnv G0)) T))
      | (ebound G0 T) => (ebound (tshiftEnv c12 G0) (tshiftTy (weakenCutoffTyVar c12 (domainEnv G0)) T))
    end.
  Fixpoint substEnv (d47 : (Trace TmVar)) (s18 : Tm) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (substEnv d47 s18 G0) T)
      | (ebound G0 T) => (ebound (substEnv d47 s18 G0) T)
    end.
  Fixpoint tsubstEnv (d47 : (Trace TyVar)) (S41 : Ty) (G : Env) : Env :=
    match G with
      | (empty) => empty
      | (evar G0 T) => (evar (tsubstEnv d47 S41 G0) (tsubstTy (weakenTrace d47 (domainEnv G0)) S41 T))
      | (ebound G0 T) => (ebound (tsubstEnv d47 S41 G0) (tsubstTy (weakenTrace d47 (domainEnv G0)) S41 T))
    end.
  Lemma domainEnv_shiftEnv  :
    (forall (c12 : (Cutoff TmVar)) (G : Env) ,
      ((domainEnv (shiftEnv c12 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainEnv_tshiftEnv  :
    (forall (c12 : (Cutoff TyVar)) (G : Env) ,
      ((domainEnv (tshiftEnv c12 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvShiftEnv .
  Qed.
  Lemma domainEnv_substEnv  :
    (forall (d47 : (Trace TmVar)) (s18 : Tm) (G : Env) ,
      ((domainEnv (substEnv d47 s18 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
  Lemma domainEnv_tsubstEnv  :
    (forall (d47 : (Trace TyVar)) (S41 : Ty) (G : Env) ,
      ((domainEnv (tsubstEnv d47 S41 G)) =
      (domainEnv G))).
  Proof.
    needleGenericDomainEnvSubstEnv .
  Qed.
End Context.
 Hint Rewrite domainEnv_appendEnv : interaction.
 Hint Rewrite domainEnv_appendEnv : env_domain_append.
Section ContextStuff.
  Section ShiftEnvAppendEnv.
    Lemma shiftEnv_appendEnv  :
      (forall (c12 : (Cutoff TmVar)) (G : Env) (G0 : Env) ,
        ((shiftEnv c12 (appendEnv G G0)) =
        (appendEnv (shiftEnv c12 G) (shiftEnv (weakenCutoffTmVar c12 (domainEnv G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
    Lemma tshiftEnv_appendEnv  :
      (forall (c12 : (Cutoff TyVar)) (G : Env) (G0 : Env) ,
        ((tshiftEnv c12 (appendEnv G G0)) =
        (appendEnv (tshiftEnv c12 G) (tshiftEnv (weakenCutoffTyVar c12 (domainEnv G)) G0)))).
    Proof.
      needleGenericShiftEnvAppendEnv .
    Qed.
  End ShiftEnvAppendEnv.
  Section SubstEnvAppendEnv.
    Lemma substEnv_appendEnv  :
      (forall (d47 : (Trace TmVar)) (s18 : Tm) (G : Env) (G0 : Env) ,
        ((substEnv d47 s18 (appendEnv G G0)) =
        (appendEnv (substEnv d47 s18 G) (substEnv (weakenTrace d47 (domainEnv G)) s18 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
    Lemma tsubstEnv_appendEnv  :
      (forall (d47 : (Trace TyVar)) (S41 : Ty) (G : Env) (G0 : Env) ,
        ((tsubstEnv d47 S41 (appendEnv G G0)) =
        (appendEnv (tsubstEnv d47 S41 G) (tsubstEnv (weakenTrace d47 (domainEnv G)) S41 G0)))).
    Proof.
      needleGenericSubstEnvAppendEnv .
    Qed.
  End SubstEnvAppendEnv.
  Section Lookups.
    Inductive lookup_evar : Env -> (Index TmVar) -> Ty -> Prop :=
      | lookup_evar_here {G : Env}
          {T98 : Ty} :
          (wfTy (domainEnv G) T98) -> (lookup_evar (evar G T98) I0 T98)
      | lookup_evar_there_evar
          {G : Env} {x52 : (Index TmVar)}
          {T98 : Ty} {T99 : Ty} :
          (lookup_evar G x52 T98) -> (lookup_evar (evar G T99) (IS x52) T98)
      | lookup_evar_there_ebound
          {G : Env} {x52 : (Index TmVar)}
          {T98 : Ty} {T99 : Ty} :
          (lookup_evar G x52 T98) -> (lookup_evar (ebound G T99) x52 (tshiftTy C0 T98)).
    Inductive lookup_ebound : Env -> (Index TyVar) -> Ty -> Prop :=
      | lookup_ebound_here {G : Env}
          {T98 : Ty} :
          (wfTy (domainEnv G) T98) -> (lookup_ebound (ebound G T98) I0 (tshiftTy C0 T98))
      | lookup_ebound_there_evar
          {G : Env} {X43 : (Index TyVar)}
          {T98 : Ty} {T99 : Ty} :
          (lookup_ebound G X43 T98) -> (lookup_ebound (evar G T99) X43 T98)
      | lookup_ebound_there_ebound
          {G : Env} {X43 : (Index TyVar)}
          {T98 : Ty} {T99 : Ty} :
          (lookup_ebound G X43 T98) -> (lookup_ebound (ebound G T99) (IS X43) (tshiftTy C0 T98)).
    Lemma lookup_evar_inversion_here  :
      (forall (G : Env) (S41 : Ty) (S42 : Ty) ,
        (lookup_evar (evar G S41) I0 S42) -> (S41 =
        S42)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_ebound_inversion_here  :
      (forall (G : Env) (S41 : Ty) (S42 : Ty) ,
        (lookup_ebound (ebound G S41) I0 S42) -> ((tshiftTy C0 S41) =
        S42)).
    Proof.
      needleGenericLookupInversion .
    Qed.
    Lemma lookup_evar_functional  :
      (forall {G : Env} {x52 : (Index TmVar)} ,
        (forall {S41 : Ty} ,
          (lookup_evar G x52 S41) -> (forall {S42 : Ty} ,
            (lookup_evar G x52 S42) -> (S41 =
            S42)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_ebound_functional  :
      (forall {G : Env} {X43 : (Index TyVar)} ,
        (forall {S41 : Ty} ,
          (lookup_ebound G X43 S41) -> (forall {S42 : Ty} ,
            (lookup_ebound G X43 S42) -> (S41 =
            S42)))).
    Proof.
      needleGenericLookupFunctional .
    Qed.
    Lemma lookup_evar_wf  :
      (forall {G : Env} {x52 : (Index TmVar)} {S41 : Ty} ,
        (lookup_evar G x52 S41) -> (wfTy (domainEnv G) S41)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma lookup_ebound_wf  :
      (forall {G : Env} {X43 : (Index TyVar)} {S41 : Ty} ,
        (lookup_ebound G X43 S41) -> (wfTy (domainEnv G) S41)).
    Proof.
      needleGenericLookupWellformedData .
    Qed.
    Lemma weaken_lookup_evar  :
      (forall {G : Env} (G0 : Env) {x52 : (Index TmVar)} {T98 : Ty} ,
        (lookup_evar G x52 T98) -> (lookup_evar (appendEnv G G0) (weakenIndexTmVar x52 (domainEnv G0)) (weakenTy T98 (domainEnv G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma weaken_lookup_ebound  :
      (forall {G : Env} (G0 : Env) {X43 : (Index TyVar)} {T98 : Ty} ,
        (lookup_ebound G X43 T98) -> (lookup_ebound (appendEnv G G0) (weakenIndexTyVar X43 (domainEnv G0)) (weakenTy T98 (domainEnv G0)))).
    Proof.
      needleGenericWeakenLookup .
    Qed.
    Lemma lookup_evar_wfindex  :
      (forall {G : Env} {x52 : (Index TmVar)} {S41 : Ty} ,
        (lookup_evar G x52 S41) -> (wfindex (domainEnv G) x52)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
    Lemma lookup_ebound_wfindex  :
      (forall {G : Env} {X43 : (Index TyVar)} {S41 : Ty} ,
        (lookup_ebound G X43 S41) -> (wfindex (domainEnv G) X43)).
    Proof.
      needleGenericLookupWellformedIndex .
    Qed.
  End Lookups.
  Inductive shift_evar : (Cutoff TmVar) -> Env -> Env -> Prop :=
    | shift_evar_here {G : Env}
        {T98 : Ty} :
        (shift_evar C0 G (evar G T98))
    | shiftevar_there_evar
        {c12 : (Cutoff TmVar)} {G : Env}
        {G0 : Env} {T : Ty} :
        (shift_evar c12 G G0) -> (shift_evar (CS c12) (evar G T) (evar G0 T))
    | shiftevar_there_ebound
        {c12 : (Cutoff TmVar)} {G : Env}
        {G0 : Env} {T : Ty} :
        (shift_evar c12 G G0) -> (shift_evar c12 (ebound G T) (ebound G0 T)).
  Lemma weaken_shift_evar  :
    (forall (G : Env) {c12 : (Cutoff TmVar)} {G0 : Env} {G1 : Env} ,
      (shift_evar c12 G0 G1) -> (shift_evar (weakenCutoffTmVar c12 (domainEnv G)) (appendEnv G0 G) (appendEnv G1 G))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Inductive shift_ebound : (Cutoff TyVar) -> Env -> Env -> Prop :=
    | shift_ebound_here {G : Env}
        {T98 : Ty} :
        (shift_ebound C0 G (ebound G T98))
    | shiftebound_there_evar
        {c12 : (Cutoff TyVar)} {G : Env}
        {G0 : Env} {T : Ty} :
        (shift_ebound c12 G G0) -> (shift_ebound c12 (evar G T) (evar G0 (tshiftTy c12 T)))
    | shiftebound_there_ebound
        {c12 : (Cutoff TyVar)} {G : Env}
        {G0 : Env} {T : Ty} :
        (shift_ebound c12 G G0) -> (shift_ebound (CS c12) (ebound G T) (ebound G0 (tshiftTy c12 T))).
  Lemma weaken_shift_ebound  :
    (forall (G : Env) {c12 : (Cutoff TyVar)} {G0 : Env} {G1 : Env} ,
      (shift_ebound c12 G0 G1) -> (shift_ebound (weakenCutoffTyVar c12 (domainEnv G)) (appendEnv G0 G) (appendEnv G1 (tshiftEnv c12 G)))).
  Proof.
    needleGenericWeakenInsertEnv .
  Qed.
  Lemma shift_evar_shifthvl_TmVar  :
    (forall {c12 : (Cutoff TmVar)} {G : Env} {G0 : Env} ,
      (shift_evar c12 G G0) -> (shifthvl_TmVar c12 (domainEnv G) (domainEnv G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Lemma shift_ebound_shifthvl_TyVar  :
    (forall {c12 : (Cutoff TyVar)} {G : Env} {G0 : Env} ,
      (shift_ebound c12 G G0) -> (shifthvl_TyVar c12 (domainEnv G) (domainEnv G0))).
  Proof.
    needleGenericInsertEnvInsertHvl .
  Qed.
  Inductive subst_evar (G : Env) (T98 : Ty) (s18 : Tm) : (Trace TmVar) -> Env -> Env -> Prop :=
    | subst_evar_here :
        (subst_evar G T98 s18 X0 (evar G T98) G)
    | subst_evar_there_evar
        {d47 : (Trace TmVar)} {G0 : Env}
        {G1 : Env} {T : Ty} :
        (subst_evar G T98 s18 d47 G0 G1) -> (subst_evar G T98 s18 (XS TmVar d47) (evar G0 T) (evar G1 T))
    | subst_evar_there_ebound
        {d47 : (Trace TmVar)} {G0 : Env}
        {G1 : Env} {T : Ty} :
        (subst_evar G T98 s18 d47 G0 G1) -> (subst_evar G T98 s18 (XS TyVar d47) (ebound G0 T) (ebound G1 T)).
  Lemma weaken_subst_evar {G : Env} {T98 : Ty} {s18 : Tm} :
    (forall (G0 : Env) {d47 : (Trace TmVar)} {G1 : Env} {G2 : Env} ,
      (subst_evar G T98 s18 d47 G1 G2) -> (subst_evar G T98 s18 (weakenTrace d47 (domainEnv G0)) (appendEnv G1 G0) (appendEnv G2 G0))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Inductive subst_ebound (G : Env) (T98 : Ty) (S41 : Ty) : (Trace TyVar) -> Env -> Env -> Prop :=
    | subst_ebound_here :
        (subst_ebound G T98 S41 X0 (ebound G T98) G)
    | subst_ebound_there_evar
        {d47 : (Trace TyVar)} {G0 : Env}
        {G1 : Env} {T : Ty} :
        (subst_ebound G T98 S41 d47 G0 G1) -> (subst_ebound G T98 S41 (XS TmVar d47) (evar G0 T) (evar G1 (tsubstTy d47 S41 T)))
    | subst_ebound_there_ebound
        {d47 : (Trace TyVar)} {G0 : Env}
        {G1 : Env} {T : Ty} :
        (subst_ebound G T98 S41 d47 G0 G1) -> (subst_ebound G T98 S41 (XS TyVar d47) (ebound G0 T) (ebound G1 (tsubstTy d47 S41 T))).
  Lemma weaken_subst_ebound {G : Env} {T98 : Ty} {S41 : Ty} :
    (forall (G0 : Env) {d47 : (Trace TyVar)} {G1 : Env} {G2 : Env} ,
      (subst_ebound G T98 S41 d47 G1 G2) -> (subst_ebound G T98 S41 (weakenTrace d47 (domainEnv G0)) (appendEnv G1 G0) (appendEnv G2 (tsubstEnv d47 S41 G0)))).
  Proof.
    needleGenericWeakenSubstEnv .
  Qed.
  Lemma subst_evar_substhvl_TmVar {G : Env} {s18 : Tm} {T98 : Ty} :
    (forall {d47 : (Trace TmVar)} {G0 : Env} {G1 : Env} ,
      (subst_evar G T98 s18 d47 G0 G1) -> (substhvl_TmVar (domainEnv G) d47 (domainEnv G0) (domainEnv G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
  Lemma subst_ebound_substhvl_TyVar {G : Env} {S41 : Ty} {T99 : Ty} :
    (forall {d47 : (Trace TyVar)} {G0 : Env} {G1 : Env} ,
      (subst_ebound G T99 S41 d47 G0 G1) -> (substhvl_TyVar (domainEnv G) d47 (domainEnv G0) (domainEnv G1))).
  Proof.
    needleGenericSubstEnvSubstHvl .
  Qed.
End ContextStuff.
 Hint Rewrite domainEnv_tshiftEnv : interaction.
 Hint Rewrite domainEnv_tshiftEnv : env_domain_shift.
 Hint Rewrite domainEnv_tsubstEnv : interaction.
 Hint Rewrite domainEnv_tsubstEnv : env_domain_subst.
 Hint Rewrite tshiftEnv_appendEnv : interaction.
 Hint Rewrite tshiftEnv_appendEnv : env_shift_append.
 Hint Rewrite tsubstEnv_appendEnv : interaction.
 Hint Rewrite tsubstEnv_appendEnv : env_shift_append.
 Hint Constructors lookup_evar lookup_ebound : infra.
 Hint Constructors lookup_evar lookup_ebound : lookup.
 Hint Constructors lookup_evar lookup_ebound : shift.
 Hint Constructors lookup_evar lookup_ebound : subst.
 Hint Resolve weaken_lookup_evar weaken_lookup_ebound : infra.
 Hint Resolve weaken_lookup_evar weaken_lookup_ebound : lookup.
 Hint Resolve weaken_lookup_evar weaken_lookup_ebound : shift.
 Hint Resolve weaken_lookup_evar weaken_lookup_ebound : weaken.
Lemma weaken_lookup_evar_here  :
  (forall {G : Env} (G0 : Env) {T98 : Ty} (wf : (wfTy (domainEnv G) T98)) ,
    (lookup_evar (appendEnv (evar G T98) G0) (weakenIndexTmVar I0 (domainEnv G0)) (weakenTy T98 (domainEnv G0)))).
Proof.
  eauto with lookup .
Qed.
Lemma weaken_lookup_ebound_here  :
  (forall {G : Env} (G0 : Env) {T98 : Ty} (wf : (wfTy (domainEnv G) T98)) ,
    (lookup_ebound (appendEnv (ebound G T98) G0) (weakenIndexTyVar I0 (domainEnv G0)) (weakenTy (tshiftTy C0 T98) (domainEnv G0)))).
Proof.
  eauto with lookup .
Qed.
 Hint Constructors wfPat wfTm wfTy : infra.
 Hint Constructors wfPat wfTm wfTy : wf.
 Hint Extern 10 ((wfPat _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTm _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 10 ((wfTy _ _)) => autorewrite with env_domain_append in *  : infra wf.
 Hint Extern 2 ((wfTy _ _)) => match goal with
  | H : (wfTy _ (tvar _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (top)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (tarrow _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (tall _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTy _ (tprod _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfPat _ _)) => match goal with
  | H : (wfPat _ (pvar)) |- _ => inversion H; subst; clear H
  | H : (wfPat _ (pprod _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Extern 2 ((wfTm _ _)) => match goal with
  | H : (wfTm _ (var _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (abs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (app _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (tabs _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (tapp _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (prod _ _)) |- _ => inversion H; subst; clear H
  | H : (wfTm _ (lett _ _ _)) |- _ => inversion H; subst; clear H
end : infra wf.
 Hint Resolve lookup_evar_wf lookup_ebound_wf : infra.
 Hint Resolve lookup_evar_wf lookup_ebound_wf : wf.
 Hint Resolve lookup_evar_wfindex lookup_ebound_wfindex : infra.
 Hint Resolve lookup_evar_wfindex lookup_ebound_wfindex : lookup.
 Hint Resolve lookup_evar_wfindex lookup_ebound_wfindex : wf.
 Hint Constructors shift_evar shift_ebound : infra.
 Hint Constructors shift_evar shift_ebound : shift.
 Hint Constructors shift_evar shift_ebound : subst.
 Hint Resolve weaken_shift_evar weaken_shift_ebound : infra.
 Hint Resolve weaken_shift_evar weaken_shift_ebound : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_ebound_shifthvl_TyVar : infra.
 Hint Resolve shift_evar_shifthvl_TmVar shift_ebound_shifthvl_TyVar : shift.
 Hint Resolve shift_evar_shifthvl_TmVar shift_ebound_shifthvl_TyVar : shift_wf.
 Hint Resolve shift_evar_shifthvl_TmVar shift_ebound_shifthvl_TyVar : wf.
 Hint Constructors subst_evar subst_ebound : infra.
 Hint Constructors subst_evar subst_ebound : subst.
 Hint Resolve weaken_subst_evar weaken_subst_ebound : infra.
 Hint Resolve weaken_subst_evar weaken_subst_ebound : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_ebound_substhvl_TyVar : infra.
 Hint Resolve subst_evar_substhvl_TmVar subst_ebound_substhvl_TyVar : subst.
 Hint Resolve subst_evar_substhvl_TmVar subst_ebound_substhvl_TyVar : subst_wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_ebound_substhvl_TyVar : wf.
 Hint Resolve subst_evar_substhvl_TmVar subst_ebound_substhvl_TyVar : substenv_substhvl.
Lemma shift_evar_lookup_evar  :
  (forall {c12 : (Cutoff TmVar)} {G : Env} {G0 : Env} (ins : (shift_evar c12 G G0)) {x52 : (Index TmVar)} {T98 : Ty} ,
    (lookup_evar G x52 T98) -> (lookup_evar G0 (shiftIndex c12 x52) T98)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_evar_lookup_ebound  :
  (forall {c12 : (Cutoff TmVar)} {G : Env} {G0 : Env} (ins : (shift_evar c12 G G0)) {X43 : (Index TyVar)} {T98 : Ty} ,
    (lookup_ebound G X43 T98) -> (lookup_ebound G0 X43 T98)).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_ebound_lookup_evar  :
  (forall {c12 : (Cutoff TyVar)} {G : Env} {G0 : Env} (ins : (shift_ebound c12 G G0)) {x52 : (Index TmVar)} {T98 : Ty} ,
    (lookup_evar G x52 T98) -> (lookup_evar G0 x52 (tshiftTy c12 T98))).
Proof.
  needleGenericInsertLookup .
Qed.
Lemma shift_ebound_lookup_ebound  :
  (forall {c12 : (Cutoff TyVar)} {G : Env} {G0 : Env} (ins : (shift_ebound c12 G G0)) {X43 : (Index TyVar)} {T98 : Ty} ,
    (lookup_ebound G X43 T98) -> (lookup_ebound G0 (tshiftIndex c12 X43) (tshiftTy c12 T98))).
Proof.
  needleGenericInsertLookup .
Qed.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_ebound shift_ebound_lookup_evar shift_ebound_lookup_ebound : infra.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_ebound shift_ebound_lookup_evar shift_ebound_lookup_ebound : lookup.
 Hint Resolve shift_evar_lookup_evar shift_evar_lookup_ebound shift_ebound_lookup_evar shift_ebound_lookup_ebound : shift.
Lemma subst_evar_lookup_ebound {G : Env} {T100 : Ty} {s18 : Tm} :
  (forall {d47 : (Trace TmVar)} {G0 : Env} {G1 : Env} (sub : (subst_evar G T100 s18 d47 G0 G1)) {X43 : (Index TyVar)} {T101 : Ty} ,
    (lookup_ebound G0 X43 T101) -> (lookup_ebound G1 X43 T101)).
Proof.
  needleGenericSubstEnvLookup .
Qed.
Lemma subst_ebound_lookup_evar {G : Env} {T100 : Ty} {S41 : Ty} (wf : (wfTy (domainEnv G) S41)) :
  (forall {d47 : (Trace TyVar)} {G0 : Env} {G1 : Env} (sub : (subst_ebound G T100 S41 d47 G0 G1)) {x52 : (Index TmVar)} {T101 : Ty} ,
    (lookup_evar G0 x52 T101) -> (lookup_evar G1 x52 (tsubstTy d47 S41 T101))).
Proof.
  needleGenericSubstEnvLookup .
Qed.
 Hint Resolve subst_evar_lookup_ebound subst_ebound_lookup_evar : infra.
 Hint Resolve subst_evar_lookup_ebound subst_ebound_lookup_evar : lookup.
 Hint Resolve subst_evar_lookup_ebound subst_ebound_lookup_evar : subst.
Fixpoint size_Ty (S0 : Ty) {struct S0} : nat :=
  match S0 with
    | (tvar X) => 1
    | (top) => 1
    | (tarrow T1 T2) => (plus 1 (plus (size_Ty T1) (size_Ty T2)))
    | (tall T0 T3) => (plus 1 (plus (size_Ty T0) (size_Ty T3)))
    | (tprod T4 T5) => (plus 1 (plus (size_Ty T4) (size_Ty T5)))
  end.
Fixpoint size_Pat (p : Pat) {struct p} : nat :=
  match p with
    | (pvar) => 1
    | (pprod p1 p2) => (plus 1 (plus (size_Pat p1) (size_Pat p2)))
  end.
Fixpoint size_Tm (s : Tm) {struct s} : nat :=
  match s with
    | (var x) => 1
    | (abs T t) => (plus 1 (plus (size_Ty T) (size_Tm t)))
    | (app t1 t2) => (plus 1 (plus (size_Tm t1) (size_Tm t2)))
    | (tabs T0 t0) => (plus 1 (plus (size_Ty T0) (size_Tm t0)))
    | (tapp t3 T1) => (plus 1 (plus (size_Tm t3) (size_Ty T1)))
    | (prod t4 t5) => (plus 1 (plus (size_Tm t4) (size_Tm t5)))
    | (lett p t6 t7) => (plus 1 (plus (size_Pat p) (plus (size_Tm t6) (size_Tm t7))))
  end.
Lemma tshift_size_Ty  :
  (forall (S40 : Ty) (c9 : (Cutoff TyVar)) ,
    ((size_Ty (tshiftTy c9 S40)) =
    (size_Ty S40))).
Proof.
  apply_mutual_ind ind_Ty.
  - intros X38 c9.
  reflexivity.
  - intros c10; simpl; apply ((f_equal S)); reflexivity.
  - intros T86 IHT86 T87 IHT87.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT86 c10)).
  + exact ((IHT87 c10)).
  - intros T88 IHT88 T89 IHT89.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT88 c10)).
  + exact ((IHT89 (CS c10))).
  - intros T90 IHT90 T91 IHT91.
  intros c10; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT90 c10)).
  + exact ((IHT91 c10)).
Qed.

Lemma shift_size_Tm  :
  (forall (s18 : Tm) (c10 : (Cutoff TmVar)) ,
    ((size_Tm (shiftTm c10 s18)) =
    (size_Tm s18))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros x48 c10.
  reflexivity.
  - intros T92 t152 IHt152.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt152 (CS c11))).
  - intros t153 IHt153 t154 IHt154.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt153 c11)).
  + exact ((IHt154 c11)).
  - intros T93 t155 IHt155.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt155 c11)).
  - intros t156 IHt156 T94.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt156 c11)).
  + reflexivity.
  - intros t157 IHt157 t158 IHt158.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt157 c11)).
  + exact ((IHt158 c11)).
  - intros p16 t159 IHt159 t160 IHt160.
  intros c11; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt159 c11)).
  + exact ((IHt160 (weakenCutoffTmVar c11 (bindPat p16)))).
Qed.

Lemma tshift_size_Tm  :
  (forall (s18 : Tm) (c11 : (Cutoff TyVar)) ,
    ((size_Tm (tshiftTm c11 s18)) =
    (size_Tm s18))).
Proof.
  apply_mutual_ind ind_Tm.
  - intros X41 c11.
  reflexivity.
  - intros T95 t161 IHt161.
  pose proof ((tshift_size_Ty T95)) as IHT95.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT95 c12)).
  + exact ((IHt161 c12)).
  - intros t162 IHt162 t163 IHt163.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt162 c12)).
  + exact ((IHt163 c12)).
  - intros T96 t164 IHt164.
  pose proof ((tshift_size_Ty T96)) as IHT96.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHT96 c12)).
  + exact ((IHt164 (CS c12))).
  - intros t165 IHt165 T97.
  pose proof ((tshift_size_Ty T97)) as IHT97.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt165 c12)).
  + exact ((IHT97 c12)).
  - intros t166 IHt166 t167 IHt167.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + exact ((IHt166 c12)).
  + exact ((IHt167 c12)).
  - intros p17 t168 IHt168 t169 IHt169.
  intros c12; simpl; apply ((f_equal S)); repeat (apply ((f_equal2 plus))).
  + reflexivity.
  + exact ((IHt168 c12)).
  + exact ((IHt169 (weakenCutoffTyVar c12 (bindPat p17)))).
Qed.

 Hint Rewrite shift_size_Tm tshift_size_Tm tshift_size_Ty : interaction.
 Hint Rewrite shift_size_Tm tshift_size_Tm tshift_size_Ty : shift_size.
Lemma weaken_size_Pat  :
  (forall (k60 : Hvl) (p18 : Pat) ,
    ((size_Pat (weakenPat p18 k60)) =
    (size_Pat p18))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Tm  :
  (forall (k60 : Hvl) (s18 : Tm) ,
    ((size_Tm (weakenTm s18 k60)) =
    (size_Tm s18))).
Proof.
  needleGenericWeakenSize .
Qed.
Lemma weaken_size_Ty  :
  (forall (k60 : Hvl) (S40 : Ty) ,
    ((size_Ty (weakenTy S40 k60)) =
    (size_Ty S40))).
Proof.
  needleGenericWeakenSize .
Qed.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : interaction.
 Hint Rewrite weaken_size_Pat weaken_size_Tm weaken_size_Ty : weaken_size.